import numpy as np
from typing import Union
from pysimenv.core.base import DynSystem
from pysimenv.common.model import FlatEarthEnv


class FixedTimeFaultEstimator(DynSystem):
    """
    Reference: L. Guo, "Fixed-Time Observer Based Safety Control for a Quadrotor UAV",
    IEEE Transactions on Aerospace and Electronic Systems, 2021.
    """
    def __init__(self, initial_states: Union[list, tuple],
                 alpha: float, beta: float, k_1: float, k_2: float,
                 m: float, J: np.ndarray, **kwargs):
        super(FixedTimeFaultEstimator, self).__init__(
            initial_states={'z_1': initial_states[0], 'z_2': initial_states[1]}, **kwargs
        )
        self.alpha = alpha
        self.beta = beta
        self.k_1 = k_1
        self.k_2 = k_2
        self.m = m
        self.J = J.copy()

    def _deriv(self, z_1: np.ndarray, z_2: np.ndarray, x: np.ndarray, eta: np.ndarray, u: np.ndarray):
        """
        :param z_1: estimation for the state
        :param z_2: estimation for the actuator fault
        :param x: actual state (v_z, p, q, r)
        :param eta: Euler angles (phi, theta, psi)
        :param f_s: Commanded thrust force
        :return:
        """
        phi, theta = eta[0], eta[1]
        p, q, r = x[1:4]
        J_x, J_y, J_z = self.J[0, 0], self.J[1, 1], self.J[2, 2]
        f = np.array([
            FlatEarthEnv.grav_accel,
            (J_y - J_z)/J_x*q*r,
            (J_z - J_x)/J_y*p*r,
            (J_x - J_y)/J_z*p*q
        ])
        B = np.diag([
            -np.cos(phi)*np.cos(theta)/self.m, 1./J_x, 1./J_y, 1./J_z
        ])

        e_d = z_1 - x
        z_1_dot = -self.k_1*(
            np.power(np.abs(e_d), self.alpha)*np.sign(e_d) + np.power(np.abs(e_d), self.beta)*np.sign(e_d)) +\
            f + z_2 + B.dot(u)
        z_2_dot = -self.k_2*(
            np.power(np.abs(e_d), 2*self.alpha - 1)*np.sign(e_d) + np.power(np.abs(e_d), 2*self.beta - 1)*np.sign(e_d)
        )
        return {'z_1': z_1_dot, 'z_2': z_2_dot}

    @property
    def delta_hat(self) -> np.ndarray:
        return self.state('z_2')



import numpy as np
from typing import Union, List
from pysimenv.core.base import SimObject
from pysimenv.core.simulator import Simulator
from pysimenv.multicopter.base import MulticopterDyn, EffectorModel, Mixer


class QuadBase(SimObject):
    def __init__(self, dyn: MulticopterDyn, control: SimObject, effector: EffectorModel = None, mixer: Mixer = None, **kwargs):
        super(QuadBase, self).__init__(**kwargs)
        self.dyn = dyn
        self.control = control
        self.effector = effector
        self.mixer = mixer
        self._add_sim_objs([self.dyn, self.control])

    # implementation
    def _forward(self, **kwargs):
        """
        :param sigma_d: [x_d, y_d, z_d, phi_d]
        :param sigma_d_dot: [x_d_dot, y_d_dot, z_d_dot, phi_d_dot]
        :return:
        """
        if self.effector:
            if self.mixer:
                u_cmd = self.control.forward(self.dyn, **kwargs)  # u_cmd = {'f': total force, 'tau': moments}
                thrusts = self.mixer.convert(**u_cmd)
            else:
                thrusts = self.control.forward(self.dyn, **kwargs)  # thrusts = [f_1, f_2, ..., f_m]
            u = self.effector.convert(thrusts)
        else:
            u = self.control.forward(self.dyn, **kwargs)

        self.dyn.forward(**u)

    # implementation
    def _output(self) -> dict:
        return self.dyn.output


class QuadXEffector(EffectorModel):
    def __init__(self, d_phi: float, d_theta: float, c_tau_f: float):
        self.d_phi = d_phi
        self.d_theta = d_theta
        self.c_tau_f = c_tau_f
        self.B = np.array([
            [1., 1., 1., 1.],
            [-d_phi/2., d_phi/2., d_phi/2., -d_phi/2.],
            [d_theta/2., -d_theta/2., d_theta/2., -d_theta/2.],
            [c_tau_f, c_tau_f, -c_tau_f, -c_tau_f]
        ])  # mapping matrix

    def convert(self, f: np.ndarray) -> dict:
        u = self.B.dot(f)
        return {'f': u[0], 'tau': u[1:4]}


class QuadXMixer(Mixer):
    def __init__(self, d_phi: float, d_theta: float, c_tau_f: float):
        self.d_phi = d_phi
        self.d_theta = d_theta
        self.c_tau_f = c_tau_f
        self.B = np.array([
            [1., 1., 1., 1.],
            [-d_phi/2., d_phi/2., d_phi/2., -d_phi/2.],
            [d_theta/2., -d_theta/2., d_theta/2., -d_theta/2.],
            [c_tau_f, c_tau_f, -c_tau_f, -c_tau_f]
        ])  # mapping matrix
        self.B_inv = np.linalg.inv(self.B)

    def convert(self, f: float, tau: np.ndarray) -> np.ndarray:
        f = self.B_inv.dot(np.array([f, tau[0], tau[1], tau[2]]))
        return f


class ActuatorFault(SimObject):
    def __init__(self, t_list: List[float], alp_list: List[np.ndarray], rho_list: List[np.ndarray],
                 interval: Union[int, float] = -1):
        """
        :param t_list: [t_0, t_1, ..., t_{k-1}] time of fault occurrence
        :param alp_list: [lam_0: (M,) array, ..., lam_{k-1}] gain fault, M: the number of motors
        :param rho_list: [rho_0: (M,) array, ..., rho_{k-1}] bias fault, M: the number of motors
        :return:
        """
        super(ActuatorFault, self).__init__(interval)
        assert len(t_list) == len(alp_list), "Array sizes doesn't match."
        assert len(t_list) == len(rho_list), "Array sizes doesn't match."
        self.t_list = t_list
        self.alp_list = alp_list
        self.rho_list = rho_list

        self.alp = alp_list[0]
        self.rho = rho_list[0]
        self.next_ind = 1

    # implement
    def _forward(self, f_s: np.ndarray):
        if self.next_ind < len(self.t_list):
            if self.time >= self.t_list[self.next_ind]:
                self.alp = self.alp_list[self.next_ind]
                self.rho = self.rho_list[self.next_ind]
                self.next_ind += 1
        f_s_star = self.alp*f_s + self.rho
        return f_s_star


def main():
    model_params = {'m': 4.34, 'J': np.diag([0.0820, 0.0845, 0.1377])}

    quad_dyn = MulticopterDyn(p_0=np.array([0., 0., -1.]), v_0=np.array([1., 0., 0.]),
                              R_0=np.identity(3), omega_0=np.array([0., 0., 0.1]), **model_params,
                              name="dynamic model")
    f = 45.
    tau = np.zeros(3)

    simulator = Simulator(quad_dyn)
    simulator.propagate(dt=0.01, time=10., save_history=True, f=f, tau=tau)

    quad_dyn.default_plot()
    quad_dyn.plot_euler_angles(show=True)


if __name__ == "__main__":
    main()




import numpy as np
from pysimenv.core.base import SimObject, DynSystem
from pysimenv.common.model import PIDControl
from pysimenv.core.simulator import Simulator


class Model(SimObject):
    def __init__(self):
        super(Model, self).__init__()
        m = 1.
        c = 6.
        k = 9.8696
        A = np.array([[0., 1.], [-k/m, -c/m]])
        B = np.array([[0.], [1./m]])
        C = np.array([[1., 0.]])
        self.dyn_sys = DynSystem(
            initial_states={'x': np.zeros(2)},
            deriv_fun=lambda x, u: {'x': A.dot(x) + B.dot(u)},
            output_fun=lambda x: C.dot(x)
        )
        self.pid_control = PIDControl(k_p=50., k_d=8., k_i=45.)

        self._add_sim_objs([self.dyn_sys, self.pid_control])

    # implement
    def _forward(self, p_d):
        p = self.dyn_sys.output
        e = p_d - p
        u_pid = self.pid_control.forward(e)

        self.dyn_sys.forward(u=u_pid)


def main():
    model = Model()
    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=5., save_history=True, p_d=np.array([1.]))
    model.dyn_sys.default_plot(show=True)


if __name__ == "__main__":
    main()




import numpy as np
import matplotlib.pyplot as plt

from pysimenv.core.base import DynSystem
from pysimenv.common.model import FlatEarthEnv
from pysimenv.common import orientation


class MulticopterDyn(DynSystem):
    e3 = np.array([0., 0., 1.], dtype=np.float32)

    def __init__(self, p_0: np.ndarray, v_0: np.ndarray, R_0: np.ndarray, omega_0: np.ndarray,
                 m: float, J: np.ndarray,
                 D_v: np.ndarray = np.zeros(3), D_omega: np.ndarray = np.zeros(3), **kwargs):
        """
        :param p_0: initial position, (3,) numpy array
            v_0: initial velocity, (3,) numpy array
            R_0: initial rotation, (3, 3) numpy array
            omega_0: initial angular velocity, (3,) numpy array
        :param m: mass, float
        :param J: inertia of mass, (3, 3) numpy array
        """
        super(MulticopterDyn, self).__init__(
            initial_states={'p': p_0, 'v': v_0, 'R': R_0, 'omega': omega_0}, **kwargs
        )
        self.m = m
        self.J = J
        self.D_v = D_v
        self.D_omega = D_omega
        self.grav_accel = FlatEarthEnv.grav_accel*self.e3

        self.state_vars['R'].attach_correction_fun(orientation.correct_orthogonality)

    # implementation
    def _deriv(self, p, v, R, omega, f: float, tau: np.ndarray):
        """
        :param p: position, (3,) numpy array
        :param v: velocity, (3,) numpy array
        :param R: rotation matrix, (3, 3) numpy array
        :param omega: angular velocity, (3,) numpy array
        :param f: total thrust, float
        :param tau: moments, (3,) numpy array
        :return: [p_dot, v_dot, R_dot, omega_dot]
        """
        p_dot = v.copy()
        v_dot = self.grav_accel - 1./self.m*(f*np.dot(R, self.e3)) - 1./self.m*self.D_v.dot(v)
        R_dot = np.matmul(R, self.hat(omega))
        omega_dot = np.linalg.solve(self.J, -np.cross(omega, np.dot(self.J, omega)) + tau - self.D_omega.dot(omega))

        return {'p': p_dot, 'v': v_dot, 'R': R_dot, 'omega': omega_dot}

    # implement
    def _output(self) -> dict:
        return self.state()

    @classmethod
    def hat(cls, v):
        return np.array([
            [0, -v[2], v[1]],
            [v[2], 0, -v[0]],
            [-v[1], v[0], 0]
        ])

    @property
    def pos(self) -> np.ndarray:
        # position in the inertial frame
        return self.state('p').copy()

    @property
    def vel(self) -> np.ndarray:
        # velocity in the inertial frame
        return self.state('v').copy()

    @property
    def rotation(self) -> np.ndarray:
        # rotation matrix from the body frame to the inertial frame R_ib
        R_ib = self.state('R').copy()
        return R_ib

    @property
    def quaternion(self) -> np.ndarray:
        R_ib = self.state('R')
        q = orientation.rotation_to_quaternion(np.transpose(R_ib))
        return q

    @property
    def euler_ang(self) -> np.ndarray:
        R_ib = self.state('R')
        eta = np.array(
            orientation.rotation_to_euler_angles(np.transpose(R_ib))
        )
        return eta

    @property
    def ang_vel(self) -> np.ndarray:
        # angular velocity of the vehicle frame with respect to the inertial frame
        return self.state('omega')

    def plot_path(self, show: bool = False):
        p = self.history('p')

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.plot(p[:, 0], p[:, 1], -p[:, 2])
        ax.set_xlabel("x (m)")
        ax.set_ylabel("y (m)")
        ax.set_zlabel("h (m)")
        ax.grid()
        ax.set_title("Flight path")

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.1)

    def plot_euler_angles(self, show: bool = False):
        t = self.history('t')
        R = self.history('R')

        data_num = R.shape[0]
        euler_angle_list = np.zeros((data_num, 3))

        for i in range(data_num):
            R_bi = np.transpose(R[i, :, :])
            euler_angle_list[i, :] = orientation.rotation_to_euler_angles(R_bi)

        fig = plt.figure()
        names = ['phi', 'theta', 'psi']
        for i in range(3):
            plt.subplot(3, 1, i + 1)
            plt.plot(t, np.rad2deg(euler_angle_list[:, i]), label="Actual")
            plt.xlabel("Time (s)")
            plt.ylabel(names[i])
            plt.grid()
            plt.legend()
        plt.suptitle("Euler angles")

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)
        return fig


class EffectorModel(object):
    # to be implemented
    def convert(self, f: np.ndarray) -> dict:
        """
        :param f: thrust of each motor [f_1, f_2, ..., f_m]
        :return: {'f': float, 'tau': np.ndarray} total thrust and moments
        """
        raise NotImplementedError


class Mixer(object):
    # to be implemented
    def convert(self, f: float, tau: np.ndarray) -> np.ndarray:
        """
        :param f: total thrust
        :param tau: moments
        :return: thrust command of each motor [f_1, f_2, ..., f_m]
        """
        raise NotImplementedError





import numpy as np
from typing import Tuple, Union, Optional
from pysimenv.missile.model import PlanarVehicle
from pysimenv.common.util import distance_traj_segment


class RelKin2dim(object):
    def __init__(self, missile: PlanarVehicle, target: PlanarVehicle):
        super(RelKin2dim, self).__init__()
        self.missile = missile
        self.target = target

        self.p_rel: Optional[np.ndarray] = None
        self.v_rel: Optional[np.ndarray] = None
        self.r: Optional[float] = None
        self.los_vec: Optional[np.ndarray] = None
        self.nor_vec: Optional[np.ndarray] = None
        self.lam: Optional[float] = None
        self.omega: Optional[float] = None

        self.forward()

    def forward(self):
        p_M, v_M = self.missile.p, self.missile.v
        p_T, v_T = self.target.p, self.target.v

        self.p_rel = p_T - p_M  # relative position
        self.v_rel = v_T - v_M  # relative velocity
        self.r = np.linalg.norm(self.p_rel)  # relative distance
        self.los_vec = self.p_rel / self.r  # LOS unit vector
        self.nor_vec = np.array([-self.los_vec[1], self.los_vec[0]])  # LOS normal vector
        self.lam = np.arctan2(self.los_vec[1], self.los_vec[0])  # LOS angle

        los_vec_3 = np.array([self.los_vec[0], self.los_vec[1], 0.])
        v_rel_3 = np.array([self.v_rel[0], self.v_rel[1], 0.])
        omega_3 = np.cross(los_vec_3, v_rel_3)/self.r
        self.omega = omega_3[2]

    @property
    def v_r(self):
        if not self.v_rel:
            self.forward()
        # rel. vel. component parallel to LOS
        return np.dot(self.v_rel, self.los_vec)

    @property
    def v_lam(self):
        if not self.v_rel:
            self.forward()
        # rel. vel. component normal to LOS
        return np.dot(self.v_rel, self.nor_vec)

    @property
    def zem(self) -> float:
        z = self.r*self.v_lam / np.linalg.norm(self.v_rel)  # ZEM(Zero-effort miss)
        return z


class CloseDistCond(object):
    """
    condition for decreasing distance
    """
    def __init__(self, r_threshold: float = 10.0):
        self.r_threshold = r_threshold  # threshold for the close distance
        self.prev_r = np.inf  # previous relative distance
        self.to_stop = False  # status

    def reset(self):
        self.prev_r = np.inf
        self.to_stop = False

    def forward(self, r):
        if r < self.r_threshold:
            if r > self.prev_r + 0.01:
                self.to_stop = True
        self.prev_r = r

    def check(self) -> bool:
        return self.to_stop


def miss_distance(p_M: np.ndarray, p_T: np.ndarray, search_range: int = 1) -> float:
    num_sample = p_M.shape[0]
    d = np.linalg.norm(p_T - p_M, axis=1)
    index_c = np.argmin(d)
    d_miss = d[index_c]

    index_min = max(index_c - search_range, 0)
    index_max = min(index_c + search_range, num_sample - 1)
    for i in range(index_min, index_max):
        p0 = p_M[i, :]
        p1 = p_M[i + 1, :]
        q0 = p_T[i, :]
        q1 = p_T[i + 1, :]

        d_, _ = distance_traj_segment(p0, p1, q0, q1)
        d_miss = min(d_miss, d_)
    return d_miss


def closest_instant(p_M: np.ndarray, p_T: np.ndarray, search_range: int = 1) -> Tuple[int, float]:
    num_sample = p_M.shape[0]
    d = np.linalg.norm(p_T - p_M, axis=1)
    index_c = np.argmin(d)
    d_miss = d[index_c]

    index_min = max(index_c - search_range, 0)
    index_max = min(index_c + search_range, num_sample - 1)

    index_close = index_c
    xi_close = 0.
    for i in range(index_min, index_max):
        p0 = p_M[i, :]
        p1 = p_M[i + 1, :]
        q0 = p_T[i, :]
        q1 = p_T[i + 1, :]

        d_, xi_ = distance_traj_segment(p0, p1, q0, q1)
        if d_ < d_miss:
            index_close = i
            xi_close = xi_
            d_miss = d_

    return index_close, xi_close


def lin_interp(x_i: Union[float, np.ndarray], x_f: Union[float, np.ndarray], xi: float) -> Union[float, np.ndarray]:
    x = x_i + xi*(x_f - x_i)
    return x



import numpy as np
from pysimenv.multicopter.base import MulticopterDyn
from pysimenv.multicopter.model import QuadBase
from pysimenv.multicopter.control import FLVelControl
from pysimenv.core.simulator import Simulator


def main():
    model_params = {'m': 0.5, 'J': np.diag([4.85, 4.85, 8.81])*1e-3}
    k_p_att = np.array([1600., 1600., 1600])
    k_d_att = np.array([80., 80., 80])
    k_p_vel = np.array([5., 5., 5.])

    dyn = MulticopterDyn(p_0=np.zeros(3), v_0=np.zeros(3), R_0=np.identity(3), omega_0=np.zeros(3), **model_params,
                         name='dynamic model')
    control = FLVelControl(**model_params, k_p_att=k_p_att, k_d_att=k_d_att, k_p_vel=k_p_vel, name='controller')
    model = QuadBase(dyn=dyn, control=control, name='quad-rotor')

    simulator = Simulator(model)
    simulator.propagate(0.01, 10., True, v_d=np.array([0.5, 0., 1.]))
    model.dyn.default_plot(show=True)


if __name__ == "__main__":
    main()




import numpy as np
import scipy
from scipy.optimize import NonlinearConstraint, BFGS, SR1
from pysimenv.common.model import FlatEarthEnv


class RelaxedHoverSolver(object):
    def __init__(self, m: float, J: np.ndarray, G: np.ndarray, kappa_f: float, kappa_tau: float,
                 fault_ind: list):
        self.m = m
        self.J = J
        self.G = G  # control effectiveness matrix
        self.kappa_f = kappa_f  # propeller coefficient
        self.kappa_tau = kappa_tau  # propeller coefficient

        self.num_motor = G.shape[1]
        self.fault_ind = fault_ind

    def _ravel(self, x: np.ndarray):
        """
        :param x: [omega_mag, p, q, r, r_ij, f_i]
        where omega = [p, q, r], R_0 = (r_ij), f = (f_i)
        :return:
        """
        omega_mag = x[0]
        omega = x[1:4]
        R_0 = x[4:13].reshape((3, 3))
        f = x[13:13 + self.num_motor]
        return omega_mag, omega, R_0, f

    def _unravel(self, omega_mag: float, omega: np.ndarray, R_0: np.ndarray, f: np.ndarray):
        x = np.hstack((np.array([omega_mag]), omega, R_0.flatten(), f))
        return x

    def constraint(self, x: np.ndarray):
        """
        :param omega_mag: magnitude of the angular velocity
        :param omega: [p, q, r] angular velocity expressed in the body coordinate
        :param R_0: rotation matrix from the body frame to the inertial frame
        :param f: [f_1, f_2, .., f_p] the thrust of each motor
        :return: constraints (=0)
        """
        omega_mag, omega, R_0, f = self._ravel(x)

        omega_i = np.array([0., 0., -omega_mag])
        z_B_i = R_0[:, 2]
        u = self.G.dot(f)
        F_l, tau = u[0], u[1:4]

        c_1 = (np.matmul(R_0, R_0.transpose()) - np.identity(3)).flatten()  # orthogonality condition
        c_2 = R_0.dot(omega) - omega_i  # angular velocity relation
        c_3 = np.array([
            np.dot(z_B_i, omega_i)*F_l/self.m - FlatEarthEnv.grav_accel*omega_mag
        ])  # total thrust condition
        c_4 = np.linalg.solve(self.J, -np.cross(omega, self.J.dot(omega)) + tau)  # zero angular acceleration
        c_5 = f[self.fault_ind]  # actuator fault condition
        return np.hstack((c_1, c_2, c_3, c_4, c_5))

    def power(self, x: np.ndarray):
        f = x[13:13 + self.num_motor]

        omega_p = np.sqrt(f/self.kappa_f)
        tau_p = self.kappa_tau*np.square(omega_p)
        return np.sum(tau_p*omega_p)

    def minimum_power_solution(self):
        # initial conditions
        omega_mag = 0.
        omega = np.zeros(3)
        R_0 = np.identity(3)
        f = np.ones(self.num_motor)*self.m*FlatEarthEnv.grav_accel/self.num_motor
        x_0 = self._unravel(omega_mag, omega, R_0, f)

        nl_con = NonlinearConstraint(self.constraint, 0., 0., jac="2-point", hess=BFGS())
        res = scipy.optimize.minimize(self.power, x_0, method='trust-constr', jac="2-point", hess=SR1(),
                                    constraints=nl_con, options={'verbose': 1})
        print(res.x)







import numpy as np
import matplotlib.pyplot as plt
from pysimenv.common.model import Dyn6DOF, Dyn6DOFEuler, Dyn6DOFRotMat, SignalGenerator, Sequential, Scope
from pysimenv.core.simulator import Simulator


def main():
    m = 1.
    w, h, d = 0.3, 0.1, 0.5
    J = np.diag([1./12*m*(h**2 + d**2), 1./12*m*(w**2 + h**2), 1./12*m*(w**2 + d**2)])

    p_0 = np.zeros(3)
    v_b_0 = np.zeros(3)
    omega_0 = np.zeros(3)

    dyn6dof = Dyn6DOF(
        p_0=p_0, v_b_0=v_b_0, q_0=np.array([1., 0., 0., 0.]), omega_0=omega_0,
        m=m, J=J)
    dyn6dof_euler = Dyn6DOFEuler(
        p_0=p_0, v_b_0=v_b_0, eta_0=np.zeros(3), omega_0=omega_0, m=m, J=J
    )
    dyn6dof_rot_mat = Dyn6DOFRotMat(
        p_0=p_0, v_b_0=v_b_0, R_0=np.identity(3), omega_0=omega_0, m=m, J=J
    )

    def force_moment(t):
        f_b = np.zeros(3)
        if t < 2.5:
            m_b = np.array([0.01, 0., 0.])
        elif t < 5.:
            m_b = np.array([0., 0., 0.01])
        else:
            m_b = np.zeros(3)
        return {'f_b': f_b, 'm_b': m_b}
    source = SignalGenerator(shaping_fun=force_moment)
    scope = Scope()
    models = {
        'quaternion': Sequential(obj_list=[source, dyn6dof, scope]),
        'euler': Sequential(obj_list=[source, dyn6dof_euler, scope]),
        'rot_mat': Sequential(obj_list=[source, dyn6dof_rot_mat, scope])
    }
    line_styles = {'quaternion': '-', 'euler': '-.', 'rot_mat': ':'}
    rotations = dict()
    for key, model in models.items():
        simulator = Simulator(model)
        simulator.propagate(dt=0.01, time=20., save_history=True)
        rotations[key] = {'t': model.obj_list[2].history('t'), 'R': model.obj_list[2].history('R')}

    fig, ax = plt.subplots(3, 3)
    for i in range(3):
        for j in range(3):
            for key, data in rotations.items():
                t = data['t']
                R = data['R']
                ax[i, j].plot(t, R[:, i, j], label=key, linestyle=line_styles[key])
                ax[i, j].set_xlabel("Time (s)")
                ax[i, j].set_ylabel("R({:d},{:d})".format(i, j))
                ax[i, j].grid()
                ax[i, j].legend()
    plt.show()


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from pysimenv.common.model import Differentiator, SignalGenerator, Scope, Sequential
from pysimenv.core.simulator import Simulator


def main():
    signal_generator = SignalGenerator(lambda t: np.array([1. - np.exp(-0.2*t)]))
    derivative = Differentiator()
    scope = Scope()

    model = Sequential([
        signal_generator,
        derivative,
        scope
    ])
    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=10.)

    t = scope.history('t')
    y_numerical = scope.history('u')
    y_analytic = 0.2*np.exp(-0.2*t)

    fig, ax = plt.subplots()
    ax.plot(t, y_numerical, label="Numerical")
    ax.plot(t, y_analytic, label="Analytic")
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Derivative")
    ax.grid()
    ax.legend()
    plt.show()


if __name__ == "__main__":
    main()




import numpy as np
from pysimenv.common.model import Dyn6DOF, SignalGenerator, Sequential
from pysimenv.core.simulator import Simulator


def main():
    m = 1.
    w, h, d = 0.3, 0.1, 0.5
    J = np.diag([1./12*m*(h**2 + d**2), 1./12*m*(w**2 + h**2), 1./12*m*(w**2 + d**2)])

    dyn6dof = Dyn6DOF(
        p_0=np.zeros(3), v_b_0=np.zeros(3),
        q_0=np.array([1., 0., 0., 0.]), omega_0=np.zeros(3),
        m=m, J=J)

    def force_moment(t):
        if t < 2.:
            f_b = np.array([0., 0., 1.])
            m_b = np.array([0.015, 0., 0.])
        else:
            f_b = np.zeros(3)
            m_b = np.zeros(3)
        return {'f_b': f_b, 'm_b': m_b}
    source = SignalGenerator(shaping_fun=force_moment)
    model = Sequential(obj_list=[source, dyn6dof])
    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=5., save_history=True)
    dyn6dof.default_plot(show=True)


if __name__ == "__main__":
    main()



import numpy as np
from pysimenv.multicopter.base import MulticopterDyn
from pysimenv.multicopter.model import QuadBase
from pysimenv.multicopter.control import QuaternionAttControl
from pysimenv.common.orientation import euler_angles_to_rotation
from pysimenv.core.simulator import Simulator


def main():
    model_params = {'m': 1.3, 'J': np.diag([0.0119, 0.0119, 0.0219])}

    R_vi = euler_angles_to_rotation(np.deg2rad([10., 10., 20.]))
    R_iv = np.transpose(R_vi)
    dyn = MulticopterDyn(p_0=np.array([1., 2., 0.]), v_0=np.zeros(3), R_0=R_iv, omega_0=np.array([0., 0., 1.]),
                              **model_params, name='dynamic model')

    Q = np.diag([1., 1., 1., 0.1, 0.1, 0.1])**2
    R = 1e-4*np.identity(3)
    K = QuaternionAttControl.gain(Q, R)
    att_control = QuaternionAttControl(**model_params, K=K, name='controller')
    model = QuadBase(dyn=dyn, control=att_control)

    simulator = Simulator(model)
    simulator.propagate(0.01, 10., True, q_d=np.array([1., 0., 0., 0.]))
    model.dyn.default_plot(show=True)


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.base import SimObject
from pysimenv.missile.model import PlanarMissile, PlanarVehicle
from pysimenv.missile.guidance import Guidance2dim
from pysimenv.missile.util import RelKin2dim, CloseDistCond, closest_instant, lin_interp


class Engagement2dim(SimObject):
    INTERCEPTED = 1
    MISSILE_STOP = 2
    IS_OUT_OF_VIEW = 3

    def __init__(self, missile: PlanarMissile, target: PlanarVehicle, guidance: Guidance2dim, name="model", **kwargs):
        super(Engagement2dim, self).__init__(name=name, **kwargs)
        self.missile = missile
        self.target = target
        self.guidance = guidance
        self.rel_kin = RelKin2dim(missile, target)
        self.close_dist_cond = CloseDistCond(r_threshold=10.0)

        self._add_sim_objs([self.missile, self.target, self.guidance])

    # override
    def _reset(self):
        super(Engagement2dim, self)._reset()
        self.close_dist_cond.reset()

    # implement
    def _forward(self):
        self.rel_kin.forward()
        self.close_dist_cond.forward(r=self.rel_kin.r)
        lam = self.rel_kin.lam
        sigma = self.missile.look_angle(lam)

        a_M_cmd = self.guidance.forward(self.missile, self.target, self.rel_kin)
        self.missile.forward(a_M_cmd=a_M_cmd)
        self.target.forward()

        self._logger.append(
            t=self.time, r=self.rel_kin.r, sigma=sigma, lam=lam, omega=self.rel_kin.omega
        )

    # implement
    def _check_stop_condition(self) -> bool:
        to_stop = False

        missile_stop = self.missile.check_stop_condition()
        if self.intercepted():  # probable interception
            to_stop = True
            self.flag = self.INTERCEPTED

        if missile_stop:  # stop due to the missile
            to_stop = True
            self.flag = self.MISSILE_STOP

        return to_stop

    def intercepted(self) -> bool:
        return self.close_dist_cond.check()

    def get_info(self) -> dict:
        p_M = self.missile.kin.history('p')
        p_T = self.target.kin.history('p')
        ind_c, xi_c = closest_instant(p_M, p_T)

        p_M_c = lin_interp(p_M[ind_c], p_M[ind_c + 1], xi_c)
        p_T_c = lin_interp(p_T[ind_c], p_T[ind_c + 1], xi_c)
        d_miss = np.linalg.norm(p_M_c - p_T_c)

        gamma_M = self.missile.history('gamma')
        gamma_T = self.target.history('gamma')
        gamma_M_c = lin_interp(gamma_M[ind_c], gamma_M[ind_c + 1], xi_c)
        gamma_T_c = lin_interp(gamma_T[ind_c], gamma_T[ind_c + 1], xi_c)
        gamma_imp = gamma_M_c - gamma_T_c

        t = self.missile.history('t')
        t_imp = lin_interp(t[ind_c], t[ind_c + 1], xi_c)
        return {'miss_distance': d_miss, 'impact_angle': gamma_imp, 'impact_time': t_imp}

    def report(self):
        self.missile.report()
        if self.flag == self.INTERCEPTED:
            print("[engagement] The target has been intercepted!")
        else:
            print("[engagement] The target has been missed!")

        info = self.get_info()
        print("[engagement] Miss distance: {:.6f} (m)".format(info['miss_distance']))
        print("[engagement] Impact angle: {:.2f} (deg)".format(np.rad2deg(info['impact_angle'])))
        print("[engagement] Impact time: {:.2f} (s) \n".format(info['impact_time']))

    def plot_path(self, show=False):
        fig_ax = self.missile.plot_path(label='missile')
        self.target.plot_path(fig_ax=fig_ax, label='target', show=show)

    def plot_rel_kin(self, show=False):
        fig_axs = dict()

        t = self.history('t')
        r = self.history('r')
        sigma = self.history('sigma')
        lam = self.history('lam')
        omega = self.history('omega')
        
        fig, ax = plt.subplots(4, 1, figsize=(6, 8))
        ax[0].set_title("Rel. dist")
        ax[0].plot(t[:-1], r[:-1], label="Rel. dist")
        ax[0].set_xlabel("Time (s)")
        ax[0].set_ylabel("r (m)")
        ax[0].grid()

        ax[1].set_title("Look angle")
        ax[1].plot(t[:-1], np.rad2deg(sigma[:-1]), label="look angle")
        ax[1].set_xlabel("Time (s)")
        ax[1].set_ylabel("sigma (deg)")
        ax[1].grid()

        ax[2].set_title("LOS angle")
        ax[2].plot(t[:-1], np.rad2deg(lam[:-1]), label="LOS angle")
        ax[2].set_xlabel("Time (s)")
        ax[2].set_ylabel("lambda (deg)")
        ax[2].grid()

        ax[3].set_title("LOS rate")
        ax[3].plot(t[:-1], np.rad2deg(omega[:-1]), label="LOS rate")
        ax[3].set_xlabel("Time (s)")
        ax[3].set_ylabel("omega (deg/s)")
        ax[3].grid()
        fig.tight_layout()
        fig_axs['Rel. Kin.'] = {'fig': fig, 'ax': ax}

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)

        return fig_axs





import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PitchDyn, PlanarMissileWithPitch, PlanarMovingTarget
from pysimenv.missile.control import ThreeLoopAP
from pysimenv.missile.guidance import PurePNG2dim
from pysimenv.missile.engagement import Engagement2dim


def main():
    aero_params = dict(L_alp=1270., L_delta=80., M_alp=-74., M_q=-5., M_delta=160.)

    pitch_dyn = PitchDyn(x_0=np.zeros(3), V=300., **aero_params)
    pitch_ap = ThreeLoopAP(V=300., **aero_params, tau_d=0.15, omega_d=30., zeta_d=1.)
    missile = PlanarMissileWithPitch(p_0=[-5000., 0.], V_0=300., gamma_0=np.deg2rad(30.),
                                     pitch_dyn=pitch_dyn, pitch_ap=pitch_ap, tau=0.02, name="missile")
    target = PlanarMovingTarget(p_0=[0., 0.], V_0=20., gamma_0=0., name="target")

    guidance = PurePNG2dim(N=3.0)
    model = Engagement2dim(missile, target, guidance)
    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=30., save_history=True)

    model.report()
    model.plot_path()
    pitch_dyn.plot(show=True)


if __name__ == "__main__":
    main()



import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.multicopter.base import MulticopterDyn
from pysimenv.multicopter.model import QuadBase
from pysimenv.multicopter.control import BSControl


def main():
    model_params = {'m': 1.023, 'J': np.diag([9.5, 9.5, 1.86])*1e-3}
    alpha = np.array([16., 14., 16., 14., 16., 14., 2.5, 0.5, 2.5, 0.5, 2.5, 0.5])

    dyn = MulticopterDyn(p_0=np.zeros(3), v_0=np.zeros(3), R_0=np.identity(3), omega_0=np.zeros(3), **model_params,
                         name='dynamic model')
    control = BSControl(alpha=alpha, **model_params, name='controller')
    model = QuadBase(dyn=dyn, control=control, name='quad-rotor')

    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=10., save_history=True, sigma_d=np.array([2., 2., -2., np.deg2rad(15.)]))
    model.dyn.plot_path()
    model.dyn.plot_euler_angles()
    model.dyn.default_plot(show=True)


if __name__ == "__main__":
    main()



import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarMissile, PlanarMovingTarget
from pysimenv.missile.guidance import PurePNG2dim
from pysimenv.missile.engagement import Engagement2dim


def main():
    missile = PlanarMissile(p_0=[-5000., 3000.], V_0=300., gamma_0=np.deg2rad(-15.), name="missile")
    target = PlanarMovingTarget(p_0=[0., 0.], V_0=20., gamma_0=0., name="target")

    guidance = PurePNG2dim(N=3.0)
    model = Engagement2dim(missile, target, guidance)
    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=30., save_history=True)

    simulator.save_sim_log('./data/planar_png/')
    model.report()
    model.plot_path()
    model.plot_rel_kin(show=True)


if __name__ == "__main__":
    main()



import numpy as np
from typing import Union, Tuple
from scipy.optimize import minimize_scalar


ArrayType = Union[list, tuple, np.ndarray]


def wrap_to_pi(angle: Union[float, np.ndarray]):
    if isinstance(angle, np.ndarray):
        angle_w = angle.copy()
        angle_w[angle > np.pi] = angle[angle > np.pi] - 2*np.pi
        angle_w[angle < -np.pi] = angle[angle < -np.pi] + 2*np.pi
    else:
        angle_w = angle
        if angle_w > np.pi:
            angle_w -= 2*np.pi
        elif angle_w < -np.pi:
            angle_w += 2*np.pi
    return angle_w


def distance_traj_segment(p0: ArrayType, p1: ArrayType, q0: ArrayType, q1: ArrayType) -> Tuple[float, float]:
    """
    A function finding the distance between two trajectory segements
    :param p0: the initial point of trajectory segment 1
    :param p1: the final point of trajectory segment 1
    :param q0: the initial point of trajectory segment 2
    :param q1: the final point of trajectory segment 2
    :return: distance
    """
    p0_ = np.array(p0)
    p1_ = np.array(p1)
    q0_ = np.array(q0)
    q1_ = np.array(q1)

    def sq_distance(xi: float):
        p = p0_ + xi * (p1_ - p0_)
        q = q0_ + xi * (q1_ - q0_)
        return np.sum((p - q)**2)
    res = minimize_scalar(sq_distance, bounds=(0, 1), method='bounded')
    d = np.sqrt(res.fun)
    xi_opt = res.x
    return d, xi_opt



import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.util import SimClock, Logger


def main():
    dt = 0.01
    sim_clock = SimClock()
    logger = Logger()

    A = np.array(
        [[1., dt], [0., 1.]]
    )
    B = np.array([[0.], [dt]])
    x = np.array([0., 0.])
    u = np.array([1.])

    for i in range(100):
        logger.append(time=sim_clock.time, state=x, control=u)
        x = A.dot(x) + B.dot(u)
        sim_clock.elapse(dt)

    logged_data = logger.get()
    fig, ax = plt.subplots()
    ax.plot(logged_data['time'], logged_data['state'], label={"Pos. [m]", "Vel. [m/s]"})
    ax.set_xlabel("Time [s]")
    ax.set_ylabel("State")
    ax.grid()
    ax.legend()
    plt.show()


if __name__ == "__main__":
    main()



import numpy as np
import scipy.linalg
from typing import Tuple
from pysimenv.core.base import ArrayType


def check_orthogonality(M: np.ndarray) -> bool:
    diff = np.linalg.norm(np.matmul(M, np.transpose(M)) - np.identity(3))
    is_orthogonal = (diff < 1e-6)
    return is_orthogonal


def correct_orthogonality(M: np.ndarray) -> np.ndarray:
    sq_M = np.matmul(M, np.transpose(M))
    sqrt_M = np.real(scipy.linalg.sqrtm(sq_M))
    return np.linalg.solve(sqrt_M, M)


def check_skew_symmetry(M: np.ndarray) -> bool:
    diff = np.linalg.norm(M + np.transpose(M))
    is_skew_symmetric = (diff < 1e-4)
    return is_skew_symmetric


def correct_skew_symmetry(M: np.ndarray) -> np.ndarray:
    return (M - np.transpose(M))/2


def hat(v: np.ndarray) -> np.ndarray:
    return np.array([
        [0., -v[2], v[1]],
        [v[2], 0., -v[0]],
        [-v[1], v[0], 0.]
    ])


def basic_rotation(axis: str, phi: float) -> np.ndarray:
    """
    Suppose that the body frame {b} is rotated from
    the inertial frame {i} by phi with respect to x-axis.
    Then the rotation matrix R_{bi} from {i} to {b} can be
    expressed as
    R_{bi} = np.array([
        [1, 0, 0],
        [0, cos(phi), sin(phi)],
        [0, -sin(phi), cos(phi)]])
    and p_{b} = R_{bi}p_{i}
    :param axis: the axis of rotation, one of 'x', 'y', 'z'
    :param phi: the angle of rotation, scalar
    :return: rotation matrix, (3, 3) numpy array
    """
    c_phi = np.cos(phi)
    s_phi = np.sin(phi)

    if axis == 'x':
        return np.array([
            [1, 0, 0],
            [0, c_phi, s_phi],
            [0, -s_phi, c_phi]
        ], dtype=np.float32)
    elif axis == 'y':
        return np.array([
            [c_phi, 0, -s_phi],
            [0, 1, 0],
            [s_phi, 0, c_phi]
        ], dtype=np.float32)
    else:
        # when axis == 'z'
        return np.array([
            [c_phi, s_phi, 0],
            [-s_phi, c_phi, 0],
            [0, 0, 1]
        ], dtype=np.float32)


def euler_angles_to_rotation(eta: ArrayType) -> np.ndarray:
    """
    convention: 3-2-1 Euler angle
    The resulting rotation matrix is from inertial frame
    to body frame, R_{bi}
    :param eta: eulerAngles = [phi, theta, psi]
    :return: rotation matrix, (3, 3) numpy array
    """
    phi, theta, psi = eta[:]

    s_phi = np.sin(phi)
    c_phi = np.cos(phi)
    s_theta = np.sin(theta)
    c_theta = np.cos(theta)
    s_psi = np.sin(psi)
    c_psi = np.cos(psi)

    R_psi = np.array([
        [c_psi, s_psi, 0.],
        [-s_psi, c_psi, 0.],
        [0., 0., 1.]
    ])
    R_theta = np.array([
        [c_theta, 0., -s_theta],
        [0., 1., 0.],
        [s_theta, 0., c_theta]
    ])
    R_phi = np.array([
        [1., 0., 0.],
        [0., c_phi, s_phi],
        [0., -s_phi, c_phi]
    ])
    return np.matmul(R_phi, np.matmul(R_theta, R_psi))


def rotation_to_euler_angles(R: np.ndarray) -> list:
    r_11 = R[0, 0]
    r_12 = R[0, 1]
    r_13 = R[0, 2]
    r_23 = R[1, 2]
    r_33 = R[2, 2]

    if abs(r_13) < 1 - 1e-6:
        theta = np.arctan2(-r_13, np.sqrt(r_11**2 + r_12**2))
        psi = np.arctan2(r_12, r_11)
        phi = np.arctan2(r_23, r_33)
    else:
        r_21 = R[1, 0]
        r_22 = R[1, 1]
        if r_13 < 0:
            theta = np.pi/2
            psi = 0
            phi = np.arctan2(r_21, r_22)
        else:
            theta = -np.pi/2
            psi = 0
            phi = -np.arctan2(r_21, r_22)

    return [phi, theta, psi]


def axis_angle_to_quaternion(a: np.ndarray, phi: float) -> np.ndarray:
    q = np.hstack((
        np.array([np.cos(phi/2.)]),
        np.sin(phi/2.)*a
    ))
    return q


def quaternion_to_axis_angle(q: np.ndarray) -> Tuple[np.ndarray, float]:
    a = q[1:4]
    if np.linalg.norm(a) > 1e-8:
        a = a/np.linalg.norm(a)
    q_0 = max(-1., min(1., q[0]))
    phi = 2*np.arccos(q_0)
    return a, phi


def quaternion_to_rotation(q: np.ndarray) -> np.ndarray:
    eta = q[0]
    epsilon = q[1:4]
    R = (eta**2 - epsilon.dot(epsilon))*np.identity(3) + 2*np.outer(epsilon, epsilon) - 2*eta*hat(epsilon)
    return R


def rotation_to_quaternion(R: np.ndarray) -> np.ndarray:
    r_11 = R[0, 0]
    r_22 = R[1, 1]
    r_33 = R[2, 2]

    r_12 = R[0, 1]
    r_21 = R[1, 0]
    r_23 = R[1, 2]
    r_32 = R[2, 1]
    r_31 = R[2, 0]
    r_13 = R[0, 2]

    eta = 0.5*np.sqrt(max(0., 1. + np.trace(R)))
    epsilon_1 = np.sign(r_23 - r_32)*0.5*np.sqrt(
        max(0., 1. + r_11 - r_22 - r_33)
    )
    epsilon_2 = np.sign(r_31 - r_13)*0.5*np.sqrt(
        max(0., 1. - r_11 + r_22 - r_33)
    )
    epsilon_3 = np.sign(r_12 - r_21)*0.5*np.sqrt(
        max(0., 1. - r_11 - r_22 + r_33)
    )

    q = np.array([eta, epsilon_1, epsilon_2, epsilon_3])
    return q



import numpy as np
import scipy
import matplotlib.pyplot as plt
from pyquaternion import Quaternion
from pysimenv.core.base import SimObject
from pysimenv.common.model import FlatEarthEnv
from pysimenv.common.orientation import quaternion_to_axis_angle
from pysimenv.multicopter.base import MulticopterDyn


class BSControl(SimObject):
    """
    Back-stepping control
    """
    def __init__(self, alpha, m, J, **kwargs):
        super(BSControl, self).__init__(**kwargs)

        self.alpha = alpha
        self.m = m
        self.J = J

        J_x, J_y, J_z = np.diag(J)[:]
        self.a_1 = (J_y - J_z)/J_x
        self.a_3 = (J_z - J_x)/J_y
        self.a_5 = (J_x - J_y)/J_z
        self.b_1 = 1./J_x
        self.b_2 = 1./J_y
        self.b_3 = 1./J_z

    # implement
    def _forward(self, dyn: MulticopterDyn, sigma_d: np.ndarray, sigma_d_dot: np.ndarray = np.zeros(4)):
        """
        :param sigma_d: [x_d, y_d, z_d, psi_d]
        :param sigma_d_dot: [x_d_dot, y_d_dot, z_d_dot, psi_d_dot]
        :return:
        """
        alp_1, alp_2, alp_3, alp_4, alp_5, alp_6, alp_7, alp_8, alp_9, alp_10, alp_11, alp_12 = self.alpha[:]

        x_1, x_3, x_5 = dyn.euler_ang[:]  # phi, theta, psi
        x_2, x_4, x_6 = dyn.ang_vel[:]  # approximation of phi_dot, theta_dot, psi_dot
        x_9, x_11, x_7 = dyn.pos[:]  # x, y, z
        x_10, x_12, x_8 = dyn.vel[:]  # x_dot, y_dot, z_dot

        # translation subsystem control
        x_9_d, x_11_d, x_7_d = sigma_d[0:3]
        x_9_d_dot, x_11_d_dot, x_7_d_dot = sigma_d_dot[0:3]

        z_7 = x_7_d - x_7
        z_7_dot = x_7_d_dot - x_8
        z_8 = z_7_dot + alp_7*z_7

        z_9 = x_9_d - x_9
        z_9_dot = x_9_d_dot - x_10
        z_10 = z_9_dot + alp_9*z_9

        z_11 = x_11_d - x_11
        z_11_dot = x_11_d_dot - x_12
        z_12 = z_11_dot + alp_11*z_11

        u_1 = -self.m/(np.cos(x_1)*np.cos(x_3))*(
                -FlatEarthEnv.grav_accel + z_7 + alp_7*(z_8 - alp_7*z_7) + alp_8*z_8)
        u_x = -self.m/u_1*(
                z_9 + alp_9*(z_10 - alp_9*z_9) + alp_10*z_10
        )
        u_y = -self.m/u_1*(
                z_11 + alp_11*(z_12 - alp_11*z_11) + alp_12*z_12
        )

        x_5_d = sigma_d[3]
        x_1_d = np.arcsin(
            np.clip(np.sin(x_5_d)*u_x - np.cos(x_5_d)*u_y, -1., 1.))
        x_3_d = np.arcsin(
            np.clip((np.cos(x_5_d)*u_x + np.sin(x_5_d)*u_y)/np.cos(x_1_d), -1., 1.))

        # rotation subsystem control
        z_1 = x_1_d - x_1
        z_1_dot = -x_2
        z_2 = z_1_dot + alp_1*z_1

        z_3 = x_3_d - x_3
        z_3_dot = -x_4
        z_4 = z_3_dot + alp_3*z_3

        z_5 = x_5_d - x_5
        z_5_dot = -x_6
        z_6 = z_5_dot + alp_5*z_5

        u_2 = 1./self.b_1*(
                -self.a_1*x_4*x_6 + z_1 + alp_1*(z_2 - alp_1*z_1) + alp_2*z_2
        )
        u_3 = 1./self.b_2*(
                -self.a_3*x_2*x_6 + z_3 + alp_3*(z_4 - alp_3*z_3) + alp_4*z_4
        )
        u_4 = 1./self.b_3*(
                -self.a_5*x_2*x_4 + z_5 + alp_5*(z_6 - alp_5*z_5) + alp_6*z_6
        )
        return {'f': u_1, 'tau': np.array([u_2, u_3, u_4])}


class SMAttControl(SimObject):
    """
    Attitude control using sliding mode
    """
    def __init__(self, m, J, c, k, **kwargs):
        super(SMAttControl, self).__init__(**kwargs)
        self.m = m
        self.J = J
        self.c = c
        self.k = k

    # implement
    def _forward(self, dyn: MulticopterDyn, eta_d: np.ndarray, omega_d: np.ndarray = np.zeros(3),
                 omega_d_dot: np.ndarray = np.zeros(3)):
        switching_fun = lambda v: 2./np.pi*np.arctan(np.linalg.norm(v))*v

        eta = dyn.euler_ang
        omega = dyn.ang_vel

        e = eta_d - eta
        e_dot = omega_d - omega

        s = e_dot + self.c*e
        tau = np.cross(omega, self.J.dot(omega)) + self.J.dot(omega_d_dot + self.c*e_dot + self.k*switching_fun(s))

        f = self.m*FlatEarthEnv.grav_accel
        return {'f': f, 'tau': tau}

class FLVelControl(SimObject):
    """
    Feedback linearization velocity control
    Reference: H. Voos, "Nonlinear Control of a Quadrotor Micro-UAV using Feedback-Linearization",
    Proceedings of the 2009 IEEE International Conference on Mechatronics, 2009.
    """
    def __init__(self, m, J,
                 k_p_att: np.ndarray, k_d_att: np.ndarray, k_p_vel: np.ndarray, **kwargs):
        super(FLVelControl, self).__init__(**kwargs)
        self.m = m
        self.J = J  # (3, 3) array
        self.k_p_att = k_p_att  # (3,) array
        self.k_d_att = k_d_att  # (3,) array
        self.k_p_vel = k_p_vel  # (3,) array

    # implement
    def _forward(self, dyn: MulticopterDyn, v_d: np.ndarray = np.zeros(3)):
        """
        :param dyn: multi-copter dynamics
        :param v_d: desired velocity
        :return:
        """
        v = dyn.vel
        eta = dyn.euler_ang
        omega = dyn.ang_vel

        u_t_1, u_t_2, u_t_3 = self.k_p_vel*(v_d - v)[:]

        f = self.m*np.sqrt(u_t_1**2 + u_t_2**2 + (u_t_3 - FlatEarthEnv.grav_accel)**2)
        phi_d = np.arcsin(self.m/f*u_t_2)
        theta_d = np.arctan(u_t_1/(u_t_3 -FlatEarthEnv.grav_accel))

        eta_d = np.array([phi_d, theta_d, 0.])
        u_star = self.k_p_att*(eta_d - eta) - self.k_d_att*omega
        tau = self.J.dot(u_star) + np.cross(omega, self.J.dot(omega))

        return {'f': f, 'tau': tau}


class QuaternionAttControl(SimObject):
    """
    Attitude control based on quaternion
    Reference: J. Carino, H. Abaunza and P. Castillo,
    "Quadrotor Quaternion Control",
    2015 International Conference on Unmanned Aircraft Systems (ICUAS), 2015.
    """
    def __init__(self, J: np.ndarray, K: np.ndarray, **kwargs):
        super(QuaternionAttControl, self).__init__(**kwargs)
        self.J = J  # (3, 3) array
        self.K = K  # (3, 6) array

    # implement
    def _forward(self, dyn: MulticopterDyn, q_d: np.ndarray, omega_d: np.ndarray = np.zeros(3)) -> np.ndarray:
        q = dyn.quaternion
        omega = dyn.ang_vel

        a, phi = quaternion_to_axis_angle(q)
        theta = a*phi
        x = np.hstack((theta, omega))

        a_d, phi_d = quaternion_to_axis_angle(q_d)
        theta_d = a_d*phi_d
        x_d = np.hstack((theta_d, omega_d))

        u = -self.K.dot(x - x_d)
        tau = self.J.dot(u) + np.cross(omega, self.J.dot(omega))
        return tau

    @classmethod
    def gain(cls, Q: np.ndarray, R: np.ndarray) -> np.ndarray:
        A = np.zeros((6, 6))
        A[0:3, 3:6] = np.identity(3)

        B = np.zeros((6, 3))
        B[3:6, :] = np.identity(3)

        P = scipy.linalg.solve_continuous_are(A, B, Q, R)
        K = np.linalg.inv(R).dot(B.transpose().dot(P))
        return K


class QuaternionPosControl(SimObject):
    e3 = np.array([0., 0., 1.])
    n = np.array([0., 0., -1.])

    def __init__(self, m: float, K: np.ndarray, **kwargs):
        super(QuaternionPosControl, self).__init__(**kwargs)
        self.m = m
        self.K = K

    # implement
    def _forward(self, dyn: MulticopterDyn, p_d: np.ndarray, v_d: np.ndarray = np.zeros(3)):
        x = np.hstack((dyn.pos, dyn.vel))
        x_d = np.hstack((p_d, v_d))
        u = -self.K.dot(x - x_d)

        u_pd = u - FlatEarthEnv.grav_accel*self.e3
        f = self.m*np.linalg.norm(u_pd)

        r_d = np.hstack((
            np.array([self.n.dot(u_pd) + np.linalg.norm(u_pd)]),
            np.cross(self.n, u_pd)
        ))

        q_d = r_d/np.linalg.norm(r_d)

        x_dot = np.hstack((dyn.vel, u))
        u_pd_dot = -self.K.dot(x_dot)
        r_d_dot = np.hstack((
            np.array([self.n.dot(u_pd_dot) + u_pd.dot(u_pd_dot)/np.linalg.norm(u_pd)]),
            np.cross(self.n, u_pd_dot)
        ))
        q_d_dot = r_d_dot/np.linalg.norm(r_d) + r_d*(
            -r_d.dot(r_d_dot)/np.linalg.norm(r_d)**3)

        omega_d = 2*Quaternion(q_d).conjugate*Quaternion(q_d_dot)
        omega_d = omega_d.elements[1:4]
        return f, q_d, omega_d

    @classmethod
    def gain(cls, Q: np.ndarray, R: np.ndarray):
        A = np.zeros((6, 6))
        A[0:3, 3:6] = np.identity(3)

        B = np.zeros((6, 3))
        B[3:6, :] = np.identity(3)

        P = scipy.linalg.solve_continuous_are(A, B, Q, R)
        K = np.linalg.inv(R).dot(B.transpose().dot(P))
        return K


class OAFControl(SimObject):
    """
    Control in one actuator failure
    Reference: Y. Wu, K. Hu, X. M. Sun, and Y. Ma, "Nonlinear Control of Quadrotor for Fault Tolerance:
    A Total Failure of One Actuator", IEEE Transactions on Systems, Man, and Cybernetics: Systems, Vol. 51, No. 5,
    2021.
    """
    def __init__(self, m, J, D_v, D_omega, G, K_p, K_d, **kwargs):
        super(OAFControl, self).__init__(**kwargs)
        self.m = m
        self.J = J
        self.d_v = np.diag(D_v)
        self.d_omega = np.diag(D_omega)
        self.G = G
        self.K_p = K_p  # P gains for [x, y, z, phi, theta]
        self.K_d = K_d  # D gains for [x, y, z, phi, theta]

    # implement
    def _forward(self, dyn: MulticopterDyn, fault_ind: int, p_d, p_d_dot = np.zeros(3), p_d_2dot = np.zeros(3)):
        p = dyn.pos
        v = dyn.vel
        eta = dyn.euler_ang
        omega = dyn.ang_vel

        phi, theta, psi = eta[:]
        if np.abs(theta) > np.pi/2. - 1e-6:
            print("")
        assert np.abs(theta) < np.pi/2. - 1e-6, "[oaf_att_control] The attitude is near the singularity."

        c_phi, s_phi = np.cos(phi), np.sin(phi)
        c_theta, s_theta = np.cos(theta), np.sin(theta)
        c_psi, s_psi = np.cos(psi), np.sin(psi)
        t_theta = s_theta/c_theta

        # Position control (altitude)
        z, v_z = p[2], v[2]
        e_1, e_1_dot = p_d[2] - z, p_d_dot[2] - v_z
        k_p_1, k_d_1 = self.K_p[2], self.K_d[2]

        F_l = -self.m/(c_phi*c_theta)*(
                -FlatEarthEnv.grav_accel + 1./self.m*self.d_v[2]*v_z + k_p_1*e_1 + k_d_1*e_1_dot + p_d_2dot[2])

        # Position control (horizontal position)
        x, y = p[0:2]
        v_x, v_y = v[0:2]
        k_p_2, k_d_2 = self.K_p[0], self.K_d[0]
        k_p_3, k_d_3 = self.K_p[1], self.K_d[1]

        e_2, e_2_dot = p_d[0] - x, p_d_dot[0] - v_x
        e_3, e_3_dot = p_d[1] - y, p_d_dot[1] - v_y
        u_x = -self.m/F_l*(1./self.m*self.d_v[0]*v_x + k_p_2*e_2 + k_d_2*e_2_dot + p_d_2dot[0])
        u_y = -self.m/F_l*(1./self.m*self.d_v[1]*v_y + k_p_3*e_3 + k_d_3*e_3_dot + p_d_2dot[1])

        phi_d = np.arcsin(
            np.clip(s_psi*u_x - c_psi*u_y, -1., 1.)
        )
        theta_d = np.arcsin(
            np.clip((c_psi*u_x + s_psi*u_y)/np.cos(phi_d), -1., 1.)
        )
        phi_d = np.clip(phi_d, np.deg2rad(-15.), np.deg2rad(15.))
        theta_d = np.clip(theta_d, np.deg2rad(-15.), np.deg2rad(15.))
        self._logger.append(t=self.time, phi_d=phi_d, theta_d=theta_d)

        # Attitude control (phi, theta)
        H_inv = np.array([
            [1., s_phi*t_theta, c_phi*t_theta],
            [0., c_phi, -s_phi],
            [0., s_phi/c_theta, c_phi/c_theta]
        ])
        eta_dot = H_inv.dot(omega)
        phi_dot, theta_dot = eta_dot[0], eta_dot[1]

        eta_r = eta[0:2]
        eta_r_dot = eta_dot[0:2]
        e_4 = np.array([phi_d, theta_d]) - eta_r
        e_4_dot = -eta_r_dot  # phi_d_dot, theta_d_dot are approximated to be 0.
        k_p_4 = self.K_p[3:5]
        k_d_4 = self.K_d[3:5]

        M_1 = np.array([
            [1., s_phi*t_theta, c_phi*t_theta],
            [0., c_phi, -s_phi]
        ])
        M_1_dot = np.array([
            [0., c_phi*t_theta*phi_dot + (s_phi/c_theta**2)*theta_dot,
             -s_phi*t_theta*phi_dot + (c_phi/c_theta**2)*theta_dot],
            [0., -s_phi*phi_dot, -c_phi*phi_dot]
        ])

        G_r, g_psi = self.reduced_thrust_model(fault_ind=fault_ind)
        n_1 = np.linalg.solve(self.J, -np.cross(omega, self.J.dot(omega)) - self.d_omega*omega
                              + np.array([0., 0., g_psi[0]])*F_l)

        M_2 = np.matmul(np.linalg.inv(self.J), np.array([[1., 0.], [0., 1.], [g_psi[1], g_psi[2]]]))

        f = M_1_dot.dot(omega) + M_1.dot(n_1)
        Q = np.matmul(M_1, M_2)

        tau_r = np.linalg.pinv(Q, rcond=1e-8).dot(-f + k_p_4*e_4 + k_d_4*e_4_dot)
        u_r = np.array([F_l, tau_r[0], tau_r[1]])
        w_r = np.linalg.solve(G_r, u_r)

        w = np.hstack((w_r[0:fault_ind], np.array([0.]), w_r[fault_ind:]))
        return w

    def reduced_thrust_model(self, fault_ind: int):
        ind_i = [0, 1, 2]  # F_l, tau_phi, tau_theta
        ind_j = []
        for j in range(4):
            if not j == fault_ind:
                ind_j.append(j)
        G_r = self.G.take(ind_i, axis=0).take(ind_j, axis=1)
        g_psi = np.dot(self.G[3].take(ind_j), np.linalg.inv(G_r))
        return G_r, g_psi

    def plot_desired_att(self):
        t = self.history('t')
        phi_d = self.history('phi_d')
        theta_d = self.history('theta_d')

        fig, ax = plt.subplots()
        ax.plot(t, np.rad2deg(phi_d), label="phi_d")
        ax.plot(t, np.rad2deg(theta_d), label="theta_d")
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Desired angle (deg)")
        ax.set_title("Desired attitude")
        ax.grid()
        ax.legend()


class OAFAttControl(SimObject):
    """
    Attitude control in one actuator failure
    """
    def __init__(self, m, J, D_v, D_omega, G, K_p, K_d, **kwargs):
        super(OAFAttControl, self).__init__(**kwargs)
        self.m = m
        self.J = J
        self.d_v = np.diag(D_v)
        self.d_omega = np.diag(D_omega)
        self.G = G
        self.K_p = K_p
        self.K_d = K_d

    # implement
    def _forward(self, dyn: MulticopterDyn, fault_ind: int, zeta_d, zeta_d_dot = np.zeros(3), zeta_d_2dot = np.zeros(3)):
        # zeta = [phi, theta, z]
        eta = dyn.euler_ang
        phi, theta = eta[0], eta[1]
        assert np.abs(theta) < np.pi/2. - 1e-6, "[oaf_att_control] The attitude is near the singularity."

        c_phi, s_phi = np.cos(phi), np.sin(phi)
        c_theta, s_theta = np.cos(theta), np.sin(theta)
        t_theta = s_theta/c_theta

        # Altitude control
        z, v_z = dyn.pos[2], dyn.vel[2]
        e_1, e_1_dot = zeta_d[2] - z, zeta_d_dot[2] - v_z
        k_p_1, k_d_1 = self.K_p[2], self.K_d[2]
        z_d_2dot = zeta_d_2dot[2]

        F_l = -self.m/(c_phi*c_theta)*(
                -FlatEarthEnv.grav_accel + 1./self.m*self.d_v[2]*v_z + k_p_1*e_1 + k_d_1*e_1_dot + z_d_2dot)

        # Attitude control
        H_inv = np.array([
            [1., s_phi*t_theta, c_phi*t_theta],
            [0., c_phi, -s_phi],
            [0., s_phi/c_theta, c_phi/c_theta]
        ])
        omega = dyn.ang_vel
        eta_dot = H_inv.dot(omega)
        phi_dot, theta_dot = eta_dot[0], eta_dot[1]

        eta_r = np.array([phi, theta])
        eta_r_dot = np.array([phi_dot, theta_dot])
        e_3 = zeta_d[0:2] - eta_r
        e_3_dot = zeta_d_dot[0:2] - eta_r_dot
        k_p_2 = self.K_p[0:2]
        k_d_2 = self.K_d[0:2]
        eta_r_d_2dot = zeta_d_2dot[0:2]

        M_1 = np.array([
            [1., s_phi*t_theta, c_phi*t_theta],
            [0., c_phi, -s_phi]
        ])
        M_1_dot = np.array([
            [0., c_phi*t_theta*phi_dot + (s_phi/c_theta**2)*theta_dot,
             -s_phi*t_theta*phi_dot + (c_phi/c_theta**2)*theta_dot],
            [0., -s_phi*phi_dot, -c_phi*phi_dot]
        ])

        G_r, g_psi = self.reduced_thrust_model(fault_ind=fault_ind)
        n_1 = np.linalg.solve(self.J, -np.cross(omega, self.J.dot(omega)) - self.d_omega*omega
                              + np.array([0., 0., g_psi[0]])*F_l)

        M_2 = np.matmul(np.linalg.inv(self.J), np.array([[1., 0.], [0., 1.], [g_psi[1], g_psi[2]]]))

        f = M_1_dot.dot(omega) + M_1.dot(n_1)
        Q = np.matmul(M_1, M_2)

        tau_r = np.linalg.pinv(Q, rcond=1e-8).dot(
            -f + k_p_2*e_3 + k_d_2*e_3_dot + eta_r_d_2dot
        )
        u_r = np.array([F_l, tau_r[0], tau_r[1]])
        w_r = np.linalg.solve(G_r, u_r)

        w = np.hstack((w_r[0:fault_ind], np.array([0.]), w_r[fault_ind:]))
        return w

    def reduced_thrust_model(self, fault_ind: int):
        ind_i = [0, 1, 2]  # F_l, tau_phi, tau_theta
        ind_j = []
        for j in range(4):
            if not j == fault_ind:
                ind_j.append(j)
        G_r = self.G.take(ind_i, axis=0).take(ind_j, axis=1)
        g_psi = np.dot(self.G[3].take(ind_j), np.linalg.inv(G_r))
        return G_r, g_psi



from typing import Union

import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.base import SimObject, DynSystem, ArrayType
from pysimenv.common.model import FirstOrderLinSys


class PlanarKin(DynSystem):
    def __init__(self, p_0: ArrayType, v_0: ArrayType, name="kin", **kwargs):
        super(PlanarKin, self).__init__(initial_states={'p': p_0, 'v': v_0}, name=name, **kwargs)

    # implement
    def _deriv(self, p, v, a: np.ndarray):
        p_dot = v.copy()
        v_dot = a.copy()
        return {'p': p_dot, 'v': v_dot}

    @property
    def p(self) -> np.ndarray:
        return self.state('p').copy()

    @property
    def v(self) -> np.ndarray:
        return self.state('v').copy()

    @property
    def V(self) -> float:
        v = self.state('v')
        return np.linalg.norm(v)

    @property
    def gamma(self) -> float:
        v = self.state('v')
        return np.arctan2(v[1], v[0])

    @property
    def R_iv(self) -> np.ndarray:
        c_gamma = np.cos(self.gamma)
        s_gamma = np.sin(self.gamma)
        return np.array([
            [c_gamma, -s_gamma],
            [s_gamma, c_gamma]
        ])

    def vel_to_inertial(self, v: np.ndarray) -> np.ndarray:
        return self.R_iv.dot(v)

    def inertial_to_vel(self, v: np.ndarray) -> np.ndarray:
        return self.R_iv.transpose().dot(v)


class PitchDyn(DynSystem):
    def __init__(self, x_0: ArrayType, V, L_alp, L_delta, M_alp, M_q, M_delta, name="pitch_dyn", **kwargs):
        # x = [alp, theta, q]
        super(PitchDyn, self).__init__(initial_states={'x': x_0}, name=name, **kwargs)
        self.V = V
        self.L_alp = L_alp
        self.L_delta = L_delta
        self.M_alp = M_alp
        self.M_q = M_q
        self.M_delta = M_delta

    # implement
    def _deriv(self, x, delta: float):
        alp, q = x[0], x[2]

        alp_dot = q - 1./self.V*(self.L_alp*alp + self.L_delta*delta)
        theta_dot = q
        q_dot = self.M_alp*alp + self.M_q*q + self.M_delta*delta

        return {'x': np.array([alp_dot, theta_dot, q_dot])}

    # implement
    def _output(self):
        return self.state('x')

    @property
    def alp(self) -> float:
        return self.state('x')[0]

    @property
    def theta(self) -> float:
        return self.state('x')[1]

    @property
    def R_vb(self) -> np.ndarray:
        c_alp = np.cos(self.alp)
        s_alp = np.sin(self.alp)
        return np.array([
            [c_alp, -s_alp],
            [s_alp, c_alp]
        ])

    def lift_accel(self, delta):
        return self.L_alp*self.alp + self.L_delta*delta

    def body_to_vel(self, v: np.ndarray) -> np.ndarray:
        return self.R_vb.dot(v)

    def plot(self, show=False):
        t = self.history('t')
        data = np.hstack((
            np.rad2deg(self.history('x')),
            np.expand_dims(np.rad2deg(self.history('delta')), axis=1)
        ))

        y_labels = ['alpha (deg)', 'theta (deg)', 'q (deg/s)', 'delta (deg)']

        fig, ax = plt.subplots(4, 1)
        for i in range(4):
            ax[i].plot(t, data[:, i])
            ax[i].set_xlabel("Time (s)")
            ax[i].set_ylabel(y_labels[i])
            ax[i].grid()
        fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)


class PlanarVehicle(SimObject):
    def __init__(self, kin: PlanarKin, name="vehicle", **kwargs):
        super(PlanarVehicle, self).__init__(name=name, **kwargs)
        self.kin = kin
        self._add_sim_objs([self.kin])

    @property
    def p(self) -> np.ndarray:
        return self.kin.p

    @property
    def v(self) -> np.ndarray:
        return self.kin.v

    @property
    def V(self) -> float:
        return self.kin.V

    @property
    def gamma(self) -> float:
        return self.kin.gamma

    def _forward(self, *args, **kwargs) -> Union[None, np.ndarray, dict]:
        pass

    def plot_path(self, fig_ax=None, label="vehicle", show=False):
        if fig_ax:
            fig = fig_ax['fig']
            ax = fig_ax['ax']
        else:
            fig, ax = plt.subplots()
            ax.set_xlabel("p_x (m)")
            ax.set_ylabel("p_y (m)")
            ax.set_aspect('equal')
            ax.set_title("Flight Path")
            ax.grid()

        p = self.kin.history('p')

        if np.std(p[:, 0]) < 1e-2 and np.std(p[:, 1]) < 1e-2:
            ax.plot(p[0, 0], p[0, 1], marker='o', label=label)
        else:
            ax.plot(p[:, 0], p[:, 1], label=label)
        ax.legend()
        fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)
        return {'fig': fig, 'ax': ax}


class PlanarMissile(PlanarVehicle):
    FLAG_STALLED = 1
    FLAG_COLLIDED = 2

    def __init__(self, p_0: ArrayType, V_0: float, gamma_0: float, name="missile", **kwargs):
        v_0 = np.array([V_0*np.cos(gamma_0), V_0*np.sin(gamma_0)])
        kin = PlanarKin(p_0=p_0, v_0=v_0)
        super(PlanarMissile, self).__init__(kin=kin, name=name, **kwargs)
        self.fov_limit = np.inf  # Field-of-view limit
        self.acc_limit = np.array([-np.inf, np.inf])  # Acceleration limit
        self.ground_elev = -np.inf  # Ground elevation

    def look_angle(self, lam: float):
        sigma = self.kin.gamma - lam
        return sigma

    # implement
    def _forward(self, a_M_cmd: float):
        a_M = np.clip(a_M_cmd, self.acc_limit[0], self.acc_limit[1])
        a = self.kin.vel_to_inertial(np.array([0., a_M]))
        self.kin.forward(a=a)
        self._logger.append(t=self.time, V=self.V, gamma=self.gamma, a_M_cmd=a_M_cmd, a_M=a_M)

    # implement
    def _check_stop_condition(self) -> bool:
        to_stop = False
        if self.is_stalled():
            to_stop = True
            self.flag = self.FLAG_STALLED
        if self.is_collided():
            to_stop = True
            self.flag = self.FLAG_STALLED

        return to_stop

    def is_stalled(self):
        return self.kin.V < 1.  # when the speed is less than 1m/s

    def is_collided(self):
        return self.kin.p[1] < self.ground_elev

    def is_out_of_fov(self, lam: float):
        sigma = self.look_angle(lam)
        return abs(sigma) > self.fov_limit

    def plot_kin(self, show=False):
        t = self.kin.history('t')
        data = [self.kin.history('p')[:, 0],
                self.kin.history('p')[:, 1],
                self.history('V'),
                np.rad2deg(self.history('gamma'))]
        labels = ['p_x', 'p_y', 'V', 'gamma']

        fig, ax = plt.subplots(4, 1)
        for i in range(4):
            ax[i].plot(t, data[i], label=labels[i])
            ax[i].set_xlabel("Time (s)")
            ax[i].set_ylabel(labels[i])
            ax[i].grid()
        fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)

    def report(self):
        np.set_printoptions(precision=2, suppress=True)

        print("[{:s}] Position: {:.2f}(m), {:.2f}(m), Speed: {:.2f}(m/s), Flight path angle: {:.2f}(deg)".format(
            self.name, self.p[0], self.p[1], self.V, np.rad2deg(self.gamma)))

        if self.flag == self.FLAG_OPERATING:
            print("[{:s}] Status: operating \n".format(self.name))
        elif self.flag == self.FLAG_STALLED:
            print("[{:s}] Status: stalled \n".format(self.name))
        elif self.flag == self.FLAG_COLLIDED:
            print("[{:s}] Status: collided \n".format(self.name))


class PlanarMissileWithPitch(PlanarMissile):
    def __init__(self, p_0: ArrayType, V_0: float, gamma_0: float,
                 pitch_dyn: PitchDyn, pitch_ap: SimObject, tau, name="missile", **kwargs):
        super(PlanarMissileWithPitch, self).__init__(p_0, V_0, gamma_0, name=name, **kwargs)
        self.pitch_dyn = pitch_dyn
        self.pitch_ap = pitch_ap
        self.act_dyn = FirstOrderLinSys(x_0=np.array([0.]), tau=tau)
        self._add_sim_objs([self.pitch_dyn, self.act_dyn, self.pitch_ap])

    # override
    def look_angle(self, lam: float):
        sigma = self.pitch_dyn.theta - lam
        return sigma

    def _forward(self, a_M_cmd: float):
        delta = self.act_dyn.output
        q = self.pitch_dyn.output[2]
        a_L = self.pitch_dyn.lift_accel(delta)
        a_L_c = a_M_cmd  # approximation
        delta_c = self.pitch_ap.forward(q=q, a_L=a_L, a_L_c=a_L_c)

        self.act_dyn.forward(u=np.array([delta_c]))
        self.pitch_dyn.forward(delta=delta)

        a = self.pitch_dyn.body_to_vel(np.array([0., a_L]))
        a = self.kin.vel_to_inertial(a)
        self.kin.forward(a=a)
        self._logger.append(t=self.time, V=self.V, gamma=self.gamma)


class PlanarMovingTarget(PlanarVehicle):
    def __init__(self, p_0: ArrayType, V_0: float, gamma_0: float, name="target", **kwargs):
        v_0 = np.array([V_0*np.cos(gamma_0), V_0*np.sin(gamma_0)])
        kin = PlanarKin(p_0=p_0, v_0=v_0)
        super(PlanarMovingTarget, self).__init__(kin=kin, name=name, **kwargs)

    # implement
    @property
    def p(self):
        return self.kin.p

    @property
    def v(self):
        return self.kin.v

    def _forward(self):
        self.kin.forward(a=np.zeros(2))
        self._logger.append(t=self.time, V=self.kin.V, gamma=self.kin.gamma)



import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarMissile, PlanarMovingTarget
from pysimenv.missile.guidance import IACBPNG
from pysimenv.missile.engagement import Engagement2dim


def main():
    theta_M_0 = np.deg2rad(30.)
    theta_T = 0.

    missile = PlanarMissile(p_0=[0., 0.], V_0=200., gamma_0=theta_M_0, name="missile")
    target = PlanarMovingTarget(p_0=[3000., 0.], V_0=30., gamma_0=theta_T, name="target")

    theta_M_f = np.deg2rad(-40.)
    sigma_max = np.deg2rad(45.)
    bpng = IACBPNG(N=3., tau=1., theta_M_0=theta_M_0, theta_M_f=theta_M_f, theta_T=theta_T,
                   lam_0=0., sigma_max=sigma_max, b_max=0.35)
    model = Engagement2dim(missile, target, bpng)
    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=30., save_history=True)

    model.report()
    bpng.plot_bias()
    bpng.plot_bias_integral()
    model.plot_path()
    model.plot_rel_kin(show=True)


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.base import SimObject
from pysimenv.core.simulator import Simulator
from pysimenv.multicopter.base import MulticopterDyn
from pysimenv.multicopter.model import QuadXEffector, QuadXMixer, ActuatorFault
from pysimenv.multicopter.control import QuaternionPosControl, QuaternionAttControl
from pysimenv.multicopter.estimator import FixedTimeFaultEstimator
from pysimenv.common.model import FlatEarthEnv, SignalGenerator


class ISMC(SimObject):
    """
    Integral sliding mode control
    """
    def __init__(self, x_b_0: np.ndarray, N: np.ndarray, eps_1: float, eps_2: float,
                 J: np.ndarray, m: float, **kwargs):
        super(ISMC, self).__init__(**kwargs)
        self._add_state_vars(x_b=x_b_0)  # initialize the baseline state
        self.N = N.copy()
        self.eps_1 = eps_1
        self.eps_2 = eps_2
        self.J = J.copy()
        self.m = m
        self.s = np.zeros(4)
        self.u_f = np.zeros(4)

    def _forward(self, dyn: MulticopterDyn, x_d: np.ndarray, u_b: np.ndarray, delta_hat: np.ndarray) -> np.ndarray:
        """
        :param x_d: desired state
        :param x: actual state (v_z, p, q, r)
        :param eta: Euler angles (phi, theta, psi)
        :param u_b: baseline control input
        :param delta_hat: estimated uncertainty
        :return:
        """
        # update the baseline state
        phi, theta = dyn.euler_ang[0:2]
        p, q, r = dyn.ang_vel[:]

        J_x, J_y, J_z = self.J[0, 0], self.J[1, 1], self.J[2, 2]
        f = np.array([
            FlatEarthEnv.grav_accel,
            (J_y - J_z)/J_x*q*r,
            (J_z - J_x)/J_y*p*r,
            (J_x - J_y)/J_z*p*q
        ])
        B = np.diag([
            -np.cos(phi)*np.cos(theta)/self.m, 1./J_x, 1./J_y, 1./J_z
        ])
        x_b_dot = f + B.dot(u_b)
        self.state_vars['x_b'].set_deriv(deriv=x_b_dot)

        # calculate the control input
        x_b = self.state('x_b')
        s = self.N.dot(x_d - x_b)
        sigma = 2/np.pi*np.arctan(np.linalg.norm(s))*s
        self.u_f = -np.linalg.solve(np.matmul(self.N, B), self.N.dot(delta_hat) + self.eps_1*sigma + self.eps_2*s)

        self._logger.append(t=self.time, s=s)
        return self.u_f.copy()

    def plot_sliding_value(self, show=False):
        t = self.history('t')
        s = self.history('s')

        fig, ax = plt.subplots(figsize=(8, 6))
        for i in range(4):
            ax.plot(t, s[:, i], label="s_" + str(i))
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Sliding value")
        ax.grid()
        ax.legend()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)


class Model(SimObject):
    def __init__(self):
        super(Model, self).__init__()

        # Trajectory
        T = 15.  # (s)
        self.pos_trajectory = SignalGenerator(
            shaping_fun=lambda t: np.array([np.sin(2*np.pi*t/T), np.cos(2*np.pi*t/T), -1.])
        )
        self.vel_trajectory = SignalGenerator(
            shaping_fun=lambda t: np.array([
                2*np.pi/T*np.cos(2*np.pi*t/T), -2*np.pi/T*np.sin(2*np.pi*t/T), 0.
            ])
        )

        # Quadrotor dynamic model
        model_params = {'m': 1.212, 'J': np.diag([1.0, 8.2, 1.48])*0.01}

        self.dyn = MulticopterDyn(p_0=np.array([0., 1., -1.]), v_0=np.zeros(3), R_0=np.identity(3), omega_0=np.zeros(3),
                                  **model_params, name='dynamic model')

        # Quadrotor thrust model (QuadX configuration)
        d_phi = 0.15
        d_theta = 0.13
        c_tau_f = 0.02
        self.effector = QuadXEffector(d_phi, d_theta, c_tau_f)

        # Quadrotor actuator mixer
        self.mixer = QuadXMixer(d_phi, d_theta, c_tau_f)

        # Actuator fault model
        self.actuator_fault = ActuatorFault(
            t_list=[0., 12., 25., 38.],
            alp_list=[
                np.array([1., 1., 1., 1.]),
                np.array([0.7, 1., 1., 1.]),
                np.array([0.7, 1., 0.6, 1.]),
                np.array([0.7, 0.8, 0.6, 0.9])
            ],
            rho_list=[
                np.zeros(4),
                np.zeros(4),
                np.zeros(4),
                np.zeros(4)
            ]
        )

        # Baseline controller
        K_att = QuaternionAttControl.gain(
            Q=np.diag([1., 1., 1., 0.1, 0.1, 0.1])**2,
            R=1e-4*np.identity(3)
        )
        K_pos = QuaternionPosControl.gain(
            Q=np.diag([1., 1., 1., 0.1, 0.1, 0.1])**2,
            R=np.identity(3)
        )
        self.att_control = QuaternionAttControl(model_params['J'], K_att)
        self.pos_control = QuaternionPosControl(model_params['m'], K_pos)

        # Fixed-Time Fault Estimator
        v_z_0 = 0.
        p_0, q_0, r_0 = 0., 0., 0.
        z_1_0 = np.array([v_z_0, p_0, q_0, r_0])
        z_2_0 = np.zeros(4)

        self.estimator = FixedTimeFaultEstimator(
            initial_states=[z_1_0, z_2_0],
            alpha=0.733, beta=1.285, k_1=20., k_2=100., **model_params
        )

        # Integral Sliding Mode Controller
        self.ismc = ISMC(
            x_b_0=np.zeros(4), N=np.diag([1., 1., 1., 1.]), eps_1=0.5, eps_2=2.5, **model_params)

        self._add_sim_objs([
            self.pos_trajectory, self.vel_trajectory,
            self.dyn, self.actuator_fault, self.att_control, self.pos_control,
            self.estimator, self.ismc])

    def forward(self) -> None:
        p_d = self.pos_trajectory.forward()
        v_d = self.vel_trajectory.forward()

        # Baseline control (position)
        F_m, q_d, omega_d = self.pos_control.forward(self.dyn, p_d, v_d)

        # Baseline control (attitude)
        M = self.att_control.forward(self.dyn, q_d, omega_d)
        u_b = np.array([F_m, M[0], M[1], M[2]])

        # Uncertainty compensation
        v_z_d = v_d[2]
        x_d = np.array([v_z_d, omega_d[0], omega_d[1], omega_d[2]])

        delta_hat = self.estimator.delta_hat
        u_f = self.ismc.forward(self.dyn, x_d, u_b, delta_hat)
        u = u_b + u_f

        # actuator fault
        f_s = self.mixer.convert(f=u[0], tau=u[1:4])
        f_s_star = self.actuator_fault.forward(f_s)
        u_star = self.effector.convert(f_s_star)
        self.dyn.forward(**u_star)

        # Fault estimation
        v_z = self.dyn.vel[2]
        p, q, r = self.dyn.ang_vel[:]
        x = np.array([v_z, p, q, r])
        eta = self.dyn.euler_ang

        self.estimator.forward(x=x, eta=eta, u=u)

        # true uncertainty
        m = self.dyn.m
        J = self.dyn.J
        J_x, J_y, J_z = J[0, 0], J[1, 1], J[2, 2]

        phi, theta = eta[0:2]
        B = np.diag([
            -np.cos(phi)*np.cos(theta)/m, 1./J_x, 1./J_y, 1./J_z
        ])
        f_star = u_star['f']
        tau_star = u_star['tau']
        delta = B.dot(np.array([f_star, tau_star[0], tau_star[1], tau_star[2]]) - u)

        self._logger.append(t=self.time, f_s=f_s, f_s_star=f_s_star,
                            delta=delta, delta_hat=delta_hat, p_d=p_d, p=self.dyn.pos)

    def plot_actuator_log(self, show=False):
        t = self.history('t')
        f_s = self.history('f_s')
        f_s_star = self.history('f_s_star')

        fig, ax = plt.subplots(4, 1, figsize=(8, 6))
        ylabels = ["Motor 1 (N)", "Motor 2 (N)", "Motor 3 (N)", "Motor 4 (N)"]
        for i in range(4):
            ax[i].plot(t, f_s[:, i], label="Command")
            ax[i].plot(t, f_s_star[:, i], label="Actual", linestyle='--')
            ax[i].set_xlabel("Time (s)")
            ax[i].set_ylabel(ylabels[i])
            ax[i].grid()
            ax[i].legend()
        fig.suptitle("Actuator status")
        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)

    def plot_uncertainty(self, show=False):
        t = self.history('t')
        delta = self.history('delta')
        delta_hat = self.history('delta_hat')

        fig, ax = plt.subplots(4, 1, figsize=(8, 6))
        ylabels = ["Delta v_z", "Delta p", "Delta q", "Delta r"]
        for i in range(4):
            ax[i].plot(t, delta[:, i], label="Actual")
            ax[i].plot(t, delta_hat[:, i], label="Estimated", linestyle='--')
            ax[i].set_xlabel("Time (s)")
            ax[i].set_ylabel(ylabels[i])
            ax[i].grid()
            ax[i].legend()
        fig.suptitle("Uncertainty estimation")
        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)

    def plot_tracking_performance(self, show=False):
        t = self.history('t')
        p_d = self.history('p_d')
        p = self.history('p')

        fig, ax = plt.subplots(3, 1, figsize=(8, 6))
        ylabels = ["x (m)", "y (m)", "z (m)"]
        for i in range(3):
            ax[i].plot(t, p_d[:, i], label="Desired")
            ax[i].plot(t, p[:, i], label="Actual")
            ax[i].set_xlabel("Time (s)")
            ax[i].set_ylabel(ylabels[i])
            ax[i].grid()
            ax[i].legend()
        fig.suptitle("Trajectory")
        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)


def main():
    model = Model()
    simulator = Simulator(model)
    simulator.propagate(0.01, 60., True)
    # model.quadrotor_dyn.default_plot(show=False)
    model.plot_tracking_performance(show=False)
    model.plot_actuator_log(show=False)
    model.plot_uncertainty(show=True)


if __name__ == "__main__":
    main()



import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.common.orientation import euler_angles_to_rotation
from pysimenv.multicopter.base import MulticopterDyn
from pysimenv.multicopter.model import QuadBase
from pysimenv.multicopter.control import SMAttControl


def main():
    model_params = {'m': 2., 'J': np.diag([1.25, 1.25, 2.50]),
                    'D_v': np.diag([1.2, 1.2, 1.6]), 'D_omega': np.diag([0.06, 0.06, 0.1])}
    R_bi = euler_angles_to_rotation(np.deg2rad([10., 10., 20.]))
    R_ib = np.transpose(R_bi)
    dyn = MulticopterDyn(p_0=np.array([1., 0., 0.]), v_0=np.zeros(3), R_0=R_ib, omega_0=np.zeros(3),
                         **model_params, name='dynamic model')
    control = SMAttControl(**model_params, c=np.array([10., 10., 10.]), k=np.array([5., 5., 5.]))
    model = QuadBase(dyn=dyn, control=control)

    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=20., save_history=True, eta_d=np.zeros(3))
    dyn.plot_euler_angles(show=True)


if __name__ == "__main__":
    main()



import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.multicopter.base import MulticopterDyn
from pysimenv.multicopter.model import QuadXEffector, QuadBase
from pysimenv.multicopter.control import OAFControl


def main():
    model_params = {'m': 1.023, 'J': np.diag([9.5, 9.5, 1.86])*0.001,
                    'D_v': np.diag([1.2, 1.2, 1.6]), 'D_omega': np.diag([0.06, 0.06, 0.1])}

    dyn = MulticopterDyn(p_0=np.array([0., 0., -5.]), v_0=np.zeros(3), R_0=np.identity(3), omega_0=np.zeros(3), **model_params,
                         name='dynamic model')
    effector = QuadXEffector(d_phi=0.15, d_theta=0.15, c_tau_f=0.0196)
    control = OAFControl(**model_params, G=effector.B,
                            K_p=np.array([1.24, 1.24, 1.24, 225., 225.]),
                            K_d=np.array([1., 1., 1., 30., 30.]))

    model = QuadBase(dyn=dyn, control=control, effector=effector)
    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=20., save_history=True, fault_ind=2, p_d=np.array([2., 2., 0.]))
    dyn.plot_path()
    dyn.plot_euler_angles()
    control.plot_desired_att()
    dyn.default_plot(show=True)


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.base import SimObject


class ThreeLoopAP(SimObject):
    def __init__(self, V, L_alp, M_q, M_delta, tau_d, omega_d, zeta_d, name="autopilot", **kwargs):
        super(ThreeLoopAP, self).__init__(name=name, **kwargs)
        self.K_A = 1./(L_alp*tau_d) - 1./V
        self.K_DC = V/(V - L_alp*tau_d)
        self.K_R = (2*zeta_d*omega_d + M_q)/M_delta
        self.omega_i = omega_d**2/(2*zeta_d*omega_d + M_q)
        self._add_state_vars(e_q_i=np.array([0.]))

    def _forward(self, q, a_L, a_L_c) -> float:
        q_c = self.K_A*(self.K_DC*a_L_c - a_L)
        e_q_i = self.state('e_q_i')[0]
        delta_c = self.K_R*(self.omega_i*e_q_i - q)

        e_q = q_c - q
        self.state_vars['e_q_i'].set_deriv(deriv=e_q)

        self._logger.append(t=self.time, a_L=a_L, a_L_c=a_L_c)
        return delta_c

    def plot_accel(self, show=False):
        data = self.history('t', 'a_L', 'a_L_c')

        fig, ax = plt.subplots()
        ax.plot(data['t'], data['a_L'], label="Actual")
        ax.plot(data['t'], data['a_L_c'], label="Command")
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Acceleration (m/s^2)")
        ax.grid()
        ax.legend()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)



import os
import h5py
import numpy as np
import ray
import time
from typing import List
from pytictoc import TicToc
from ray.remote_function import RemoteFunction
from pysimenv.core.util import SimClock, Timer, Logger
from pysimenv.core.base import SimObject


class Simulator(object):
    def __init__(self, model: SimObject, verbose: bool = True, reset_model: bool = True):
        self.sim_clock = SimClock()
        self.log_timer = Timer(np.inf)
        self.log_timer.attach_sim_clock(self.sim_clock)

        self.model = model
        self.model.attach_sim_clock(self.sim_clock)
        self.model.attach_log_timer(self.log_timer)
        if reset_model:
            self.model.reset()

        self.state_vars = model.collect_state_vars()
        self.verbose = verbose

    def reset(self):
        self.sim_clock.reset()
        self.log_timer.turn_off()
        self.model.reset()

    def begin_logging(self, log_interval: float):
        # log_interval must be greater than or equal to dt
        self.log_timer.turn_on(log_interval)

    def finish_logging(self):
        self.log_timer.turn_off()

    def step(self, dt: float, **kwargs):
        t_0 = self.sim_clock.time

        self.sim_clock.major_time_step = True
        self.log_timer.forward()
        self.model.forward(**kwargs)
        for var in self.state_vars:
            var.rk4_update_1(dt)
        self.sim_clock.major_time_step = False

        self.sim_clock.apply_time(t_0 + dt / 2.)
        self.log_timer.forward()
        self.model.forward(**kwargs)
        for var in self.state_vars:
            var.rk4_update_2(dt)

        self.model.forward(**kwargs)
        for var in self.state_vars:
            var.rk4_update_3(dt)

        self.sim_clock.apply_time(t_0 + dt - 10 * self.sim_clock.time_res)
        self.log_timer.forward()
        self.model.forward(**kwargs)
        for var in self.state_vars:
            var.rk4_update_4(dt)

        self.sim_clock.apply_time(t_0 + dt)

    def propagate(self, dt: float, time: float, save_history: bool = True, *args, **kwargs):
        if save_history:
            self.begin_logging(dt)

        self.sim_clock.set_time_interval(dt)
        self.model.check_sim_clock()
        self.model.initialize()

        if self.verbose:
            print("[simulator] Simulating...")
        tic_toc = TicToc()
        tic_toc.tic()

        # perform propagation
        iter_num = min(round(time/dt), np.iinfo(np.int32).max)
        for i in range(iter_num):
            to_stop = self.model.check_stop_condition()
            if to_stop:
                break
            self.step(dt, **kwargs)

        self.log_timer.forward()
        self.model.forward(**kwargs)

        # self.model.propagate(dt, time, **kwargs)
        elapsed_time = tic_toc.tocvalue()

        if self.verbose:
            print("[simulator] Elapsed time: {:.4f} [s]".format(elapsed_time))
        self.finish_logging()

    def save_sim_log(self, save_dir='./data/'):
        os.makedirs(save_dir, exist_ok=True)
        file = h5py.File(save_dir + 'log.hdf5', 'w')
        self.model.save(file)
        file.close()

    def load_sim_log(self, save_dir='./data/'):
        file = h5py.File(save_dir + 'log.hdf5', 'r')
        self.model.load(file)
        file.close()


class ParallelSimulator(object):
    def __init__(self, simulation_fun: RemoteFunction):
        """
        :param simulation_fun: reference for a function object.
        """
        self.simulation_fun = simulation_fun
        self.logger = Logger()
        self.name = "par_simulator"

    def simulate(self, parameter_sets: List[dict], verbose: bool = False):
        print("[{:s}] Initializing the parallel simulation...".format(self.name))
        ray.init()

        num_sim = len(parameter_sets)
        remaining_ids = []
        parameters_mapping = {}

        for parameters in parameter_sets:
            result_id = self.simulation_fun.remote(**parameters)
            remaining_ids.append(result_id)
            parameters_mapping[result_id] = parameters

        num_done = 0
        start = time.time()
        print("[{:s}] Simulating...".format(self.name))
        while remaining_ids:
            done_ids, remaining_ids = ray.wait(remaining_ids)
            result_id = done_ids[0]
            num_done += 1

            parameters = parameters_mapping[result_id]
            result = ray.get(result_id)
            self.logger.append(**parameters, **result)
            print("[{:s}] [{:d}/{:d}] ".format(self.name, num_done, num_sim))
            if verbose:
                print("Parameters: " + str(parameters))
                print("Result: " + str(result) + "\n")

        duration = round(time.time() - start)
        print("[{:s}] Simulations done in {:d} seconds".format(self.name, duration))

    def get(self, *args):
        return self.logger.get(*args)

    def save(self, save_dir=None, data_group=''):
        if save_dir is None:
            save_dir = './data/'
        os.makedirs(save_dir, exist_ok=True)
        file = h5py.File(save_dir + 'par_sim.hdf5', 'w')
        self.logger.save(file, data_group)



import numpy as np
import matplotlib.pyplot as plt
from typing import Union
from pysimenv.core.base import SimObject
from pysimenv.missile.model import PlanarMissile, PlanarVehicle
from pysimenv.missile.util import RelKin2dim


class Guidance2dim(SimObject):
    def __init__(self, interval: Union[int, float] = -1, name="guidance", **kwargs):
        super(Guidance2dim, self).__init__(interval=interval, name=name, **kwargs)

    # to be implemented
    def _forward(self, missile: PlanarMissile, target: PlanarVehicle, rel_kin: RelKin2dim) -> float:
        raise NotImplementedError


class PurePNG2dim(Guidance2dim):
    def __init__(self, N: float = 3.0, interval: Union[int, float] = -1, **kwargs):
        super(PurePNG2dim, self).__init__(interval=interval, **kwargs)
        self.N = N

    # implement
    def _forward(self, missile: PlanarMissile, target: PlanarVehicle, rel_kin: RelKin2dim) -> float:
        """
        :return: acceleration command a_y_cmd
        """
        V_M = missile.kin.V  # speed of the missile
        omega = rel_kin.omega  # LOS rate

        a_M_cmd = self.N*V_M*omega
        return a_M_cmd


class IACBPNG(Guidance2dim):
    """
    Biased PNG with terminal-angle constraint (impact angle control)
    """
    def __init__(self, N: float, tau: float, theta_M_0: float, theta_M_f: float, theta_T: float, lam_0: float,
                 sigma_max: float = float('inf'), b_max: float = float('inf'), **kwargs):
        super(IACBPNG, self).__init__(**kwargs)
        self._add_state_vars(B=np.array([0.]))

        # Guidance parameters
        self.N = N
        self.tau = tau
        self.theta_M_0 = theta_M_0
        self.theta_M_f = theta_M_f
        self.theta_T = theta_T
        self.lam_0 = lam_0
        self.sigma_max = sigma_max
        self.b_max = b_max

        self.phase = 0  # 0: BPNG, 1: DPP, 2: BPNG

    @classmethod
    def reference_value(cls, theta_M_0, theta_M_f, theta_T, lam_0, N, V_M, V_T):
        theta_M_0_bar = theta_M_0 - lam_0
        theta_M_f_bar = theta_M_f - lam_0
        theta_T_bar = theta_T - lam_0

        lam_f_bar = np.arctan(
            (V_M*np.sin(theta_M_f_bar) - V_T*np.sin(theta_T_bar))/(V_M*np.cos(theta_M_f_bar) - V_T*np.cos(theta_T_bar))
        )
        B_ref = theta_M_f_bar - theta_M_0_bar - N*lam_f_bar
        return B_ref

    def _forward(self, missile: PlanarMissile, target: PlanarVehicle, rel_kin: RelKin2dim) -> float:
        V_M = missile.V
        V_T = target.V
        omega = rel_kin.omega
        sigma = missile.look_angle(rel_kin.lam)

        B_ref = IACBPNG.reference_value(
            self.theta_M_0, self.theta_M_f, self.theta_T, self.lam_0, self.N, V_M, V_T)
        B = self.state('B')[0]

        b_1 = 1./self.tau*(B_ref - B)
        b_2 = (1. - self.N)*omega
        if self.phase == 0:
            # BPNG
            b = b_1
            if np.abs(sigma) > self.sigma_max - np.deg2rad(0.01):
                self.phase = 1
        elif self.phase == 1:
            # DPP
            b = b_2
            if np.abs(b_1) < np.abs(b_2):
                self.phase = 2
        else:
            # BPNG
            b = b_1

        b = np.clip(b, -self.b_max, self.b_max)
        a_png = self.N*V_M*omega
        a_b = V_M*b
        a_M = a_png + a_b

        self.state_vars['B'].set_deriv(deriv=b)
        self._logger.append(t=self.time, B_ref=B_ref, B=B, b=b, a_png=a_png, a_b=a_b)
        return a_M

    def plot_bias(self):
        t = self.history('t')
        b = self.history('b')

        fig, ax = plt.subplots()
        ax.plot(t, b)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Bias, b (rad/s)")
        ax.grid()

    def plot_bias_integral(self):
        t = self.history('t')
        B_ref = self.history('B_ref')
        B = self.history('B')

        fig, ax = plt.subplots()
        ax.plot(t, B_ref, linestyle='--')
        ax.plot(t, B)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Integral Value of Bias, B (rad)")
        ax.grid()



import numpy as np
from pysimenv.core.base import SimObject
from pysimenv.multicopter.base import MulticopterDyn
from pysimenv.multicopter.control import QuaternionPosControl, QuaternionAttControl
from pysimenv.core.simulator import Simulator


class Model(SimObject):
    def __init__(self):
        super(Model, self).__init__()
        model_params = {'m': 1.3, 'J': np.diag([0.0119, 0.0119, 0.0219])}

        self.dyn = MulticopterDyn(p_0=np.array([1., 2., 0.]), v_0=np.zeros(3), R_0=np.identity(3), omega_0=np.zeros(3),
                                  **model_params, name='dynamic_model')

        K_att = QuaternionAttControl.gain(
            Q=np.diag([1., 1., 1., 0.1, 0.1, 0.1])**2,
            R=1e-4*np.identity(3)
        )
        self.att_control = QuaternionAttControl(J=model_params['J'], K=K_att)

        K_pos = QuaternionPosControl.gain(
            Q=np.diag([1., 1., 1., 0.1, 0.1, 0.1])**2,
            R=np.identity(3)
        )
        self.pos_control = QuaternionPosControl(m=model_params['m'], K=K_pos)

        self._add_sim_objs([self.dyn, self.att_control, self.pos_control])

    def _forward(self, p_d: np.ndarray, v_d: np.ndarray = np.zeros(3)):
        f, q_d, omega_d = self.pos_control.forward(self.dyn, p_d, v_d)
        tau = self.att_control.forward(self.dyn, q_d, omega_d)

        self.dyn.forward(f=f, tau=tau)


def main():
    p_d = np.array([0., 0., 1.])
    v_d = np.array([0., 0., 0.])
    model = Model()
    simulator = Simulator(model)
    simulator.propagate(0.01, 10., True, p_d=p_d, v_d=v_d)
    model.dyn.default_plot(show=True)


if __name__ == "__main__":
    main()



"""
Description for Package
"""
from pysimenv.core.base import StateVariable, SimObject, StaticObject, DynSystem, TimeVaryingDynSystem, BaseFunction
from pysimenv.core.simulator import Simulator
from pysimenv.core.util import SimClock, Timer, Logger

__all__ = ['core']



from pysimenv.common.model import SignalGenerator, Integrator, Sequential
from pysimenv.core.simulator import Simulator


def main():
    signal_generator = SignalGenerator(
        shaping_fun=lambda t: 1./(1. + t)**2
    )
    integrator = Integrator([0.])
    model = Sequential([signal_generator, integrator])
    simulator = Simulator(model)
    simulator.propagate(0.01, 10., True)
    integrator.default_plot(show=True)


if __name__ == "__main__":
    main()






import numpy as np
from typing import Optional, Union
from pysimenv.core.error import NoSimClockError


class SimClock(object):
    def __init__(self, time: float = 0., time_res: float = 1e-6):
        self.time = time  # Current time
        self.time_res = time_res  # Time resolution
        self.dt = None
        self.major_time_step: bool = True

    def reset(self):
        self.time = 0.
        self.dt = None

    def apply_time(self, time: float):
        self.time = time

    def set_time_res(self, time_res: float):
        self.time_res = time_res

    def set_time_interval(self, dt: float):
        self.dt = dt

    def elapse(self, dt: float):
        self.time += dt


class Timer(object):
    def __init__(self, event_time_interval: Union[int, float] = -1):
        # construction phase
        self.sim_clock: Optional[SimClock] = None
        self.event_time_interval = event_time_interval

        self.is_operating = False
        self.is_event = False
        self._last_event_time = -np.inf

    def reset(self):
        self.turn_on()
        self._last_event_time = -np.inf

    def attach_sim_clock(self, sim_clock: SimClock):
        # initialization phase
        self.sim_clock = sim_clock

    def turn_on(self, event_time_interval: Union[None, int, float] = None):
        # initialization phase
        if self.sim_clock is None:
            raise NoSimClockError()
        if event_time_interval is not None:
            self.event_time_interval = event_time_interval

        self.is_operating = True
        self.is_event = True
        self._last_event_time = self.sim_clock.time

    def turn_off(self):
        self.is_operating = False
        self.is_event = False
        self._last_event_time = -np.inf

    def detach_sim_clock(self):
        self.sim_clock = None

    def forward(self):
        if self.is_operating:
            # prevent repetitive calling
            if abs(self.sim_clock.time - self._last_event_time) <= self.sim_clock.time_res:
                return

            # always raise an event when the event time interval is -1
            if self.event_time_interval == -1:
                self.is_event = True
                self._last_event_time = self.sim_clock.time
                return

            elapsed_time = self.sim_clock.time - self._last_event_time
            if elapsed_time >= (self.event_time_interval - self.sim_clock.time_res):
                self.is_event = True
                self._last_event_time = self.sim_clock.time
            else:
                self.is_event = False

    @staticmethod
    def test():
        print("== Test for Timer ==")

        dt = 0.01
        sim_clock = SimClock()

        event_time_interval = 0.1
        timer = Timer(event_time_interval)
        timer.attach_sim_clock(sim_clock)
        timer.turn_on()

        print("Initial time: 0.0[s]")
        print("Sampling time interval: {:.2f}[s]".format(dt))
        print("Event time interval: {:.2f}[s]\n".format(event_time_interval))
        for i in range(100):
            timer.forward()
            if timer.is_event:
                print("[Timer] An event occurred at time = {:.1f}[s]".format(sim_clock.time))
            sim_clock.elapse(dt)


class Logger:
    def __init__(self):
        self._log_timer: Optional[Timer] = None
        self.data = dict()
        self.is_operating = True

    def attach_log_timer(self, log_timer: Timer):
        self._log_timer = log_timer

    def turn_on(self):
        self.is_operating = True

    def turn_off(self):
        self.is_operating = False

    def detach_log_timer(self):
        self._log_timer = None

    def clear(self):
        self.data = dict()

    def is_empty(self):
        return (self.data is None) or (len(self.data) == 0)

    def len(self):
        if self.data is None:
            return 0
        first_key = next(iter(self.data.keys()))
        return len(self.data[first_key])

    def append(self, **kwargs):
        if self.is_operating:
            if (self._log_timer is None) or self._log_timer.is_event:
                for key in kwargs.keys():
                    arr = kwargs[key]
                    if isinstance(arr, np.ndarray):
                        arr = arr.copy()

                    try:
                        self.data[key].append(arr)
                    except KeyError:
                        self.data[key] = []
                        self.data[key].append(arr)

    def get(self, *args) -> Union[None, np.ndarray, dict]:
        if self.is_empty():
            return None

        if len(args) == 0:
            args = self.data.keys()

        if len(args) == 1:
            return np.array(self.data[args[0]])
        else:
            return {key: np.array(self.data[key]) for key in args}

    def keys(self):
        return self.data.keys()

    def save(self, h5file=None, data_group=''):
        logged_data = self.get()
        if logged_data is not None:
            for key, value in logged_data.items():
                h5file[data_group + '/' + key] = value

    def load(self, h5file=None, data_group=''):
        try:
            for key, value in h5file[data_group].items():
                self.data[key] = list(value)
        except KeyError:
            pass


if __name__ == "__main__":
    Timer.test()




"""
Description for Package
"""
from pysimenv.multicopter.base import MulticopterDyn, EffectorModel, Mixer
from pysimenv.multicopter.model import QuadBase, QuadXEffector, QuadXMixer, ActuatorFault
from pysimenv.multicopter.control import BSControl, SMAttControl, FLVelControl, QuaternionAttControl, QuaternionPosControl
from pysimenv.multicopter.estimator import FixedTimeFaultEstimator

__all__ = ['multicopter']



import matplotlib.pyplot as plt
import numpy as np
from typing import Union, List, Tuple, Optional
from pysimenv.core.base import SimObject, DynSystem, ArrayType
from pysimenv.common import orientation, util


class SignalGenerator(SimObject):
    def __init__(self, shaping_fun, interval: Union[int, float] = -1, **kwargs):
        super(SignalGenerator, self).__init__(interval=interval, **kwargs)
        self.shaping_fun = shaping_fun

    # implementation
    def _forward(self, *args, **kwargs) -> Union[np.ndarray, dict]:
        return self.shaping_fun(self.time, *args, **kwargs)


class Scope(SimObject):
    def __init__(self, **kwargs):
        super(Scope, self).__init__(**kwargs)

    # implementation
    def _forward(self, **kwargs):
        self._logger.append(t=self.time, **kwargs)

    def plot(self, show: bool = False):
        var_names = list(self._logger.keys())
        var_names.remove('t')

        t = self.history('t')
        for var_name in var_names:
            var = self.history(var_name)

            fig, ax = plt.subplots()
            for i in range(var.shape[1]):
                ax.plot(t, var[:, i], label=var_name + "_" + str(i))
                ax.set_xlabel("Time (s)")
                ax.set_ylabel("Value")
                ax.grid()
                ax.legend()
            fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)


class FeedbackControl(SimObject):
    def __init__(self, system: SimObject, control: SimObject, **kwargs):
        super(FeedbackControl, self).__init__(**kwargs)
        self.system = system
        self.control = control

        self._add_sim_objs([system, control])

    # implement
    def _forward(self, **kwargs) -> Union[None, np.ndarray, dict]:
        x = self.system.state()
        if isinstance(x, np.ndarray):
            u_fb = self.control.forward(x=x)
        else:
            u_fb = self.control.forward(**x, **kwargs)

        if isinstance(u_fb, np.ndarray):
            out = self.system.forward(u=u_fb)
        else:
            out = self.system.forward(**u_fb)
        return out


class OFBControl(SimObject):
    """
    Output feedback control
    """
    def __init__(self, system: SimObject, control: SimObject, **kwargs):
        super(OFBControl, self).__init__(**kwargs)
        self.system = system
        self.control = control

        self._add_sim_objs([system, control])

    # implement
    def _forward(self, **kwargs) -> Union[None, np.ndarray, dict]:
        y = self.system.output
        if isinstance(y, np.ndarray):
            u_fb = self.control.forward(y=y)
        else:
            u_fb = self.control.forward(**y, **kwargs)

        if isinstance(u_fb, np.ndarray):
            out = self.system.forward(u=u_fb)
        else:
            out = self.system.forward(**u_fb)
        return out


class Sequential(SimObject):
    def __init__(self, obj_list: Union[List[SimObject], Tuple[SimObject]], **kwargs):
        super(Sequential, self).__init__(**kwargs)
        assert len(obj_list) > 0, "[sequential] Invalid obj_list!"
        self.obj_list = list(obj_list)
        self.first_obj = obj_list[0]
        self.other_obj_list = obj_list[1:]

        self._add_sim_objs(obj_list)

    # implement
    def _forward(self, **kwargs) -> Union[None, np.ndarray, dict]:
        self.first_obj.forward(**kwargs)
        out = self.first_obj.output
        for obj in self.other_obj_list:
            if out is None:
                out = obj.forward()
            elif isinstance(out, dict):
                out = obj.forward(**out)
            else:
                out = obj.forward(u=out)

        return out


class FlatEarthEnv(object):
    grav_accel = 9.805


class LinSys(DynSystem):
    def __init__(self, x_0: ArrayType, A: np.ndarray, B: np.ndarray = None, C: np.ndarray = None, **kwargs):
        super(LinSys, self).__init__(initial_states={'x': x_0}, **kwargs)
        self.A = A
        self.B = B
        if C is None:
            C = np.eye(A.shape[0], dtype=np.float32)
        self.C = C

    # implement
    def _deriv(self, x: np.ndarray, u: Union[None, np.ndarray] = None):
        if self.B is None or u is None:
            x_dot = self.A.dot(x)
        else:
            x_dot = self.A.dot(x) + self.B.dot(u)

        return {'x': x_dot}

    # implement
    def _output(self) -> np.ndarray:
        return self.C.dot(self.state('x'))


class FirstOrderLinSys(LinSys):
    def __init__(self, x_0: ArrayType, tau: float, **kwargs):
        A = np.array([[-1./tau]])
        B = np.array([[1./tau]])

        super(FirstOrderLinSys, self).__init__(x_0, A, B, **kwargs)
        self.tau = tau

    # override
    def _output(self) -> float:
        return self.state('x')[0]


class SecondOrderLinSys(LinSys):
    def __init__(self, x_0: ArrayType, zeta: float, omega: float, **kwargs):
        A = np.array([
            [0., 1.],
            [-omega**2, -2*zeta*omega]
        ], dtype=np.float32)
        B = np.array([[0.], [omega**2]], dtype=np.float32)

        super(SecondOrderLinSys, self).__init__(x_0, A, B, **kwargs)
        self.zeta = zeta
        self.omega = omega

    # override
    def _output(self) -> float:
        return self.state('x')[0]


class Integrator(DynSystem):
    def __init__(self, initial_state: Union[ArrayType], **kwargs):
        def deriv_fun(x: np.ndarray, u: Union[float, np.ndarray]):
            if isinstance(u, float):
                u = np.array([u])
            return {'x': u}

        super(Integrator, self).__init__({'x': initial_state}, deriv_fun, **kwargs)


class Differentiator(SimObject):
    def __init__(self, **kwargs):
        super(Differentiator, self).__init__(**kwargs)
        self.u_prev = None
        self.t_prev = None

    # implement
    def _forward(self, u: np.ndarray) -> np.ndarray:
        if self.u_prev is None:
            deriv = np.zeros_like(u)
        else:
            deriv = (u - self.u_prev)/(self.time - self.t_prev)

        if self._sim_clock.major_time_step:
            self.u_prev = u.copy()
            self.t_prev = self.time

        return deriv


class PIDControl(SimObject):
    def __init__(self, k_p: Union[float, np.ndarray], k_i: Union[float, np.ndarray], k_d: Union[float, np.ndarray],
                 windup_limit=None, **kwargs):
        super(PIDControl, self).__init__(**kwargs)
        self._add_state_vars(e_i=np.zeros_like(k_p))
        self.k_p = k_p
        self.k_i = k_i
        self.k_d = k_d
        self.windup_limit = windup_limit
        if self.windup_limit:
            self.state_vars['e_i'].attach_correction_fun(self.clip_windup)

        self.e_prev = None
        self.t_prev = None

    # implement
    def _forward(self, e: np.ndarray) -> np.ndarray:
        self.state_vars['e_i'].set_deriv(deriv=e)

        e_i = self.state('e_i')
        e_d = np.zeros_like(e) if self.e_prev is None else (e - self.e_prev)/(self.time - self.t_prev)
        u_pid = self.k_p*e + self.k_i*e_i + self.k_d*e_d

        if self._sim_clock.major_time_step:
            self.e_prev = e.copy()
            self.t_prev = self.time

        return u_pid

    def clip_windup(self, e_i: np.ndarray) -> np.ndarray:
        return np.clip(e_i, self.windup_limit, -self.windup_limit)


class Dyn6DOF(DynSystem):
    def __init__(self, p_0: np.ndarray, v_b_0: np.ndarray, q_0: np.ndarray, omega_0: np.ndarray,
                 m: float, J: np.ndarray, **kwargs):
        super(Dyn6DOF, self).__init__(initial_states={'p': p_0, 'v_b': v_b_0, 'q': q_0, 'omega': omega_0}, **kwargs)
        self.m = m
        self.J = J

        def normalize(v: np.ndarray) -> np.ndarray:
            v_norm = np.linalg.norm(v)
            if v_norm < 1e-6:
                return np.zeros_like(v)
            else:
                return v/v_norm

        self.state_vars['q'].attach_correction_fun(normalize)

    @property
    def R(self) -> np.ndarray:
        # rotation matrix representing the rotation from the inertial frame to the body frame
        q = self.state('q')
        eta, epsilon = q[0], q[1:4]
        S_epsilon = orientation.hat(epsilon)
        R_bi = np.identity(3) - 2*eta*S_epsilon + 2*np.matmul(S_epsilon, S_epsilon)
        return R_bi

    @property
    def v_i(self) -> np.ndarray:
        v_b = self.state('v_b')
        R_ib = self.R.transpose()
        return R_ib.dot(v_b)

    # implement
    def _deriv(self, p, v_b, q, omega, f_b: np.ndarray, m_b: np.ndarray):
        p_dot = self.v_i
        v_b_dot = -np.cross(omega, v_b) + 1./self.m*f_b

        Q = np.array([
            [-q[1], -q[2], -q[3]],
            [q[0], -q[3], q[2]],
            [q[3], q[0], -q[1]],
            [-q[2], q[1], q[0]]
        ])
        q_dot = 0.5*Q.dot(omega)
        omega_dot = np.linalg.solve(self.J, -np.cross(omega, self.J.dot(omega)) + m_b)
        return {'p': p_dot, 'v_b': v_b_dot, 'q': q_dot, 'omega': omega_dot}

    # override
    def _output(self) -> Union[None, np.ndarray, dict]:
        out = self.state()
        out.update({'R': self.R, 'v_i': self.v_i})
        return out


class Dyn6DOFEuler(DynSystem):
    def __init__(self, p_0: np.ndarray, v_b_0: np.ndarray, eta_0: np.ndarray, omega_0: np.ndarray,
                 m: float, J: np.ndarray, **kwargs):
        super(Dyn6DOFEuler, self).__init__(initial_states={'p': p_0, 'v_b': v_b_0, 'eta': eta_0, 'omega': omega_0},
                                           **kwargs)
        self.m = m
        self.J = J
        self.name = "dyn_6dof_euler"

        self.state_vars['eta'].attach_correction_fun(util.wrap_to_pi)

    @property
    def R(self) -> np.ndarray:
        eta = self.state('eta')
        R_bi = orientation.euler_angles_to_rotation(eta)
        return R_bi

    @property
    def v_i(self) -> np.ndarray:
        v_b = self.state('v_b')
        R_ib = self.R.transpose()
        return R_ib.dot(v_b)

    # implement
    def _deriv(self, p, v_b, eta, omega, f_b: np.ndarray, m_b: np.ndarray):
        p_dot = self.v_i
        v_b_dot = -np.cross(omega, v_b) + 1./self.m*f_b

        phi, theta = eta[0:2]
        c_phi = np.cos(phi)
        s_phi = np.sin(phi)
        c_theta = np.cos(theta)
        t_theta = np.tan(theta)
        assert np.abs(theta) < np.pi/2. - 1e-6, "[{:s}] The attitude is near the singularity.".format(self.name)
        H_inv = np.array([
            [1., s_phi*t_theta, c_phi*t_theta],
            [0., c_phi, -s_phi],
            [0., s_phi/c_theta, c_phi/c_theta]
        ])
        eta_dot = H_inv.dot(omega)
        omega_dot = np.linalg.solve(self.J, -np.cross(omega, self.J.dot(omega)) + m_b)
        return {'p': p_dot, 'v_b': v_b_dot, 'eta': eta_dot, 'omega': omega_dot}

    # override
    def _output(self) -> Union[None, np.ndarray, dict]:
        out = self.state()
        out.update({'R': self.R, 'v_i': self.v_i})
        return out


class Dyn6DOFRotMat(DynSystem):
    def __init__(self, p_0: np.ndarray, v_b_0: np.ndarray, R_0: np.ndarray, omega_0: np.ndarray,
                 m: float, J: np.ndarray, **kwargs):
        super(Dyn6DOFRotMat, self).__init__(initial_states={'p': p_0, 'v_b': v_b_0, 'R': R_0, 'omega': omega_0},
                                            **kwargs)
        self.m = m
        self.J = J
        self.name = "dyn_6dof_rotation"

        self.state_vars['R'].attach_correction_fun(orientation.correct_orthogonality)

    @property
    def v_i(self) -> np.ndarray:
        v_b = self.state('v_b')
        R_ib = self.state('R').transpose()
        return R_ib.dot(v_b)

    # implement
    def _deriv(self, p, v_b, R, omega, f_b: np.ndarray, m_b: np.ndarray):
        p_dot = self.v_i
        v_b_dot = -np.cross(omega, v_b) + 1./self.m*f_b
        R_dot = -np.matmul(orientation.hat(omega), R)
        omega_dot = np.linalg.solve(self.J, -np.cross(omega, self.J.dot(omega)) + m_b)
        return {'p': p_dot, 'v_b': v_b_dot, 'R': R_dot, 'omega': omega_dot}

    # override
    def _output(self) -> Union[None, np.ndarray, dict]:
        out = self.state()
        out.update({'v_i': self.v_i})
        return out



import numpy as np
from scipy.optimize import fmin_slsqp
from pysimenv.multicopter.model import QuadXThrustModel
from pysimenv.multicopter.utils import RelaxedHoverSolver

m = 0.5
J = np.diag([2.7, 2.7, 5.2])*0.001
g = 9.807
d_phi = 0.17/np.sqrt(2.)
d_theta = 0.17/np.sqrt(2.)
kappa_f = 6.41e-6
kappa_tau = 1.1e-7
G = np.array([
    [kappa_f, kappa_f, kappa_f, kappa_f],
    [-kappa_f*d_phi, kappa_f*d_phi, kappa_f*d_phi, -kappa_f*d_phi],
    [kappa_f*d_theta, -kappa_f*d_theta, kappa_f*d_theta, -kappa_f*d_theta],
    [kappa_tau, kappa_tau, -kappa_tau, -kappa_tau]
])


def eq_const(x: np.ndarray):
    omega = x[0:3]
    v_op = x[3:5]

    v = np.array([v_op[0], v_op[1], 0., 0.])

    u = G.dot(np.square(v))
    p, q, r = omega
    J_x, J_y, J_z = np.diag(J)[:]

    c_1 = np.array([m*g - u[0]])
    c_2 = np.array([
        (J_z - J_y)*q*r - u[1],
        (J_x - J_z)*p*r - u[2],
        (J_y - J_x)*p*q - u[3]
    ])
    return np.hstack((c_1, c_2))


def ieq_const(x: np.ndarray):
    v_op = x[3:5]
    return v_op


def obj_fun(x: np.ndarray):
    omega = x[0:3]
    return np.sum(np.square(omega))


def main():
    omega_0 = np.zeros(3)
    v_op_0 = np.sqrt(np.ones(2)*m*g/2./kappa_f)
    x_0 = np.hstack((omega_0, v_op_0))

    x_opt = fmin_slsqp(obj_fun, x_0, eqcons=[eq_const], ieqcons=[ieq_const])
    print("-------")
    print(x_opt)


if __name__ == "__main__":
    main()



import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarMissile


def main():
    missile = PlanarMissile(p_0=[-5000., 0.], V_0=200., gamma_0=np.deg2rad(30.))
    simulator = Simulator(missile)
    simulator.propagate(dt=0.01, time=50., save_history=True, a_M_cmd=-5.)
    missile.plot_kin()
    missile.plot_path(show=True, label="missile")


if __name__ == "__main__":
    main()




import numpy as np
import matplotlib.pyplot as plt
from itertools import count
from typing import Union, Callable, Optional, Dict, List
from pysimenv.core.util import SimClock, Timer, Logger
from pysimenv.core.error import NoSimClockError


ArrayType = Union[list, tuple, np.ndarray]


class StateVariable(object):
    def __init__(self, state: ArrayType):
        if isinstance(state, float):
            state = np.array([state])
        self.state = np.array(state)
        self.deriv = np.zeros_like(state)
        self._rk4_buffer = []
        self.correction_fun = None

    def apply_state(self, state: ArrayType):
        if isinstance(state, float):
            state = np.array([state])
        self.state = np.array(state)

    @property
    def size(self):
        return self.state.size

    @property
    def shape(self):
        return self.state.shape

    def set_deriv(self, deriv: ArrayType):
        if isinstance(deriv, float):
            deriv = np.array([deriv])
        self.deriv = np.array(deriv)

    def rk4_update_1(self, dt: float):
        x_0 = np.copy(self.state)
        k_1 = np.copy(self.deriv)

        self._rk4_buffer.clear()
        self._rk4_buffer.append(x_0)
        self._rk4_buffer.append(k_1)
        self.state = x_0 + dt / 2 * k_1  # x = x0 + dt/2*k1
        if self.correction_fun:
            self.state = self.correction_fun(self.state)

    def rk4_update_2(self, dt: float):
        x_0 = self._rk4_buffer[0]
        k_2 = np.copy(self.deriv)

        self._rk4_buffer.append(k_2)
        self.state = x_0 + dt / 2 * k_2  # x = x0 + dt/2*k2
        if self.correction_fun:
            self.state = self.correction_fun(self.state)

    def rk4_update_3(self, dt: float):
        x_0 = self._rk4_buffer[0]
        k_3 = np.copy(self.deriv)

        self._rk4_buffer.append(k_3)
        self.state = x_0 + dt / 2 * k_3  # x = x0 + dt/2*k3
        if self.correction_fun:
            self.state = self.correction_fun(self.state)

    def rk4_update_4(self, dt: float):
        x_0, k_1, k_2, k_3 = self._rk4_buffer[:]
        k_4 = np.copy(self.deriv)

        self.state = x_0 + dt * (k_1 + 2 * k_2 + 2 * k_3 + k_4) / 6
        if self.correction_fun:
            self.state = self.correction_fun(self.state)

    def attach_correction_fun(self, correction_fun: Callable[[np.ndarray], np.ndarray]):
        self.correction_fun = correction_fun

    @staticmethod
    def test():
        state_var = StateVariable([1.])
        def deriv_fun(x): return -x
        dt = 0.01

        time_list = []
        sim_state_list = []
        t = 0
        for i in range(1000):
            time_list.append(t)
            sim_state_list.append(state_var.state)

            state_var.set_deriv(deriv_fun(state_var.state))
            state_var.rk4_update_1(dt)
            state_var.set_deriv(deriv_fun(state_var.state))
            state_var.rk4_update_2(dt)
            state_var.set_deriv(deriv_fun(state_var.state))
            state_var.rk4_update_3(dt)
            state_var.set_deriv(deriv_fun(state_var.state))
            state_var.rk4_update_4(dt)
            t = t + dt

        time_list = np.array(time_list)
        sim_state_list = np.array(sim_state_list)
        true_state_list = np.exp(-time_list)

        plt.figure()
        plt.subplot(2, 1, 1)
        plt.plot(time_list, sim_state_list[:, 0], color='k')
        plt.xlabel("Time [s]")
        plt.ylabel("State")
        plt.grid()

        plt.subplot(2, 1, 2)
        plt.plot(time_list, sim_state_list[:, 0] - true_state_list, color='k')
        plt.xlabel("Time [s]")
        plt.ylabel("Integration error")
        plt.grid()

        plt.subplots_adjust(bottom=0.1, left=0.1, top=0.9, hspace=0.5)
        plt.show()


class SimObject(object):
    FLAG_OPERATING = 0
    _ids = count(0)

    def __init__(self, interval: Union[int, float] = -1, name: str = None, **kwargs):
        self.id = next(self._ids)
        self.name = name if name else 'sim_obj_' + str(self.id)
        self.flag: int = SimObject.FLAG_OPERATING
        self.is_static = True
        self.state_vars: Dict[str, StateVariable] = dict()

        self.sim_objs: List[SimObject] = []
        self._sim_clock: Optional[SimClock] = None
        self._timer = Timer(event_time_interval=interval)
        self._logger = Logger()
        self._last_output = None

    def _add_state_vars(self, **kwargs):
        if kwargs:
            for name, initial_state in kwargs.items():
                if isinstance(initial_state, float):
                    initial_state = np.array([initial_state])
                var = StateVariable(initial_state)
                self.state_vars[name] = var
                self.is_static = False

    def _add_sim_objs(self, objs: Union['SimObject', list, tuple]):
        if isinstance(objs, SimObject):
            objs = [SimObject]

        for obj in objs:
            if not isinstance(obj, SimObject):
                continue
            if obj in self.sim_objs:
                continue
            self.sim_objs.append(obj)
            self.is_static = self.is_static and obj.is_static

    def collect_state_vars(self) -> List[StateVariable]:
        state_vars = []
        for var in self.state_vars.values():
            state_vars.append(var)
        for sim_obj in self.sim_objs:
            state_vars.extend(sim_obj.collect_state_vars())
        return state_vars

    @property
    def num_state_vars(self):
        return len(self.state_vars)

    @property
    def num_sim_objs(self):
        return len(self.sim_objs)

    def _attach_sim_clock(self, sim_clock: SimClock):
        self._sim_clock = sim_clock
        self._timer.attach_sim_clock(sim_clock)
        self._timer.turn_on()

    def _attach_log_timer(self, log_timer: Timer):
        self._logger.attach_log_timer(log_timer)

    def _initialize(self):
        pass

    def _detach_sim_clock(self):
        self._sim_clock = None
        self._timer.turn_off()
        self._timer.detach_sim_clock()

    def _detach_log_timer(self):
        self._logger.detach_log_timer()

    def _reset(self):
        self.flag = SimObject.FLAG_OPERATING
        self._timer.reset()
        self._logger.clear()

    def attach_sim_clock(self, sim_clock: SimClock):
        self._attach_sim_clock(sim_clock)
        for sim_obj in self.sim_objs:
            sim_obj.attach_sim_clock(sim_clock)

    def attach_log_timer(self, log_timer: Timer):
        self._attach_log_timer(log_timer)
        for sim_obj in self.sim_objs:
            sim_obj.attach_log_timer(log_timer)

    def initialize(self):
        self._initialize()
        for sim_obj in self.sim_objs:
            sim_obj.initialize()

    def detach_sim_clock(self):
        self._detach_sim_clock()
        for sim_obj in self.sim_objs:
            sim_obj.detach_sim_clock()

    def detach_log_timer(self):
        self._detach_log_timer()
        for sim_obj in self.sim_objs:
            sim_obj.detach_log_timer()
    
    def reset(self):
        self._reset()
        for sim_obj in self.sim_objs:
            sim_obj.reset()

    def check_sim_clock(self):
        assert self._sim_clock is not None, "Attach a sim_clock first!"
        for sim_obj in self.sim_objs:
            sim_obj.check_sim_clock()

    def set_state(self, **kwargs) -> None:
        """
        :param kwargs: states
        :return: None
        """
        if len(kwargs) > 0:
            for name, state in kwargs.items():
                self.state_vars[name].apply_state(state)

    @property
    def time(self) -> float:
        if self._sim_clock is None:
            raise NoSimClockError
        return self._sim_clock.time

    def state(self, *args) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        if len(args) == 1:
            return self.state_vars[args[0]].state
        elif len(args) == 0:
            return self._get_states()
        else:
            states = dict()
            for name in args:
                var = self.state_vars[name]
                states[name] = var.state
            return states

    def _get_states(self) -> Dict[str, np.ndarray]:
        states = dict()
        for name, var in self.state_vars.items():
            states[name] = var.state
        return states

    def deriv(self, *args) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        if len(args) == 1:
            return self.state_vars[args[0]].deriv
        elif len(args) == 0:
            return self._get_derivs()
        else:
            derivs = dict()
            for name in args:
                var = self.state_vars[name]
                derivs[name] = var.deriv
            return derivs

    def _get_derivs(self) -> Dict[str, np.ndarray]:
        derivs = dict()
        for name, var in self.state_vars.items():
            derivs[name] = var.deriv
        return derivs

    @property
    def output(self, *args, **kwargs) -> Union[None, float, np.ndarray, dict]:
        return self._last_output

    def forward(self, *args, **kwargs) -> Union[None, float, np.ndarray, dict]:
        self._timer.forward()
        if self.is_static:
            if self._timer.is_event:
                output = self._forward(*args, **kwargs)
                self._last_output = output
        else:
            output = self._forward(*args, **kwargs)
            if self._timer.is_event:
                self._last_output = output

        return self._last_output

    # should be implemented
    def _forward(self, *args, **kwargs) -> Union[None, float, np.ndarray, dict]:
        raise NotImplementedError

    # to be implemented
    def _check_stop_condition(self, *args, **kwargs) -> Optional[bool]:
        pass

    def check_stop_condition(self, *args, **kwargs) -> bool:
        to_stop_list = [self._check_stop_condition(*args, **kwargs)]
        for sim_obj in self.sim_objs:
            to_stop_list.append(sim_obj.check_stop_condition(*args, **kwargs))

        to_stop = np.array(to_stop_list, dtype=bool).any()
        return to_stop

    def history(self, *args):
        """
        :param args: variable names
        :return:
        """
        return self._logger.get(*args)

    # to be implemented
    def report(self):
        pass

    def default_plot(self, show=False, var_keys=None, var_ind_dict=None, var_names_dict=None):
        if var_keys is None:
            var_keys = list(self._logger.keys())
            var_keys.remove('t')
        if var_ind_dict is None:
            var_ind_dict = dict()
        if var_names_dict is None:
            var_names_dict = dict()

        fig_axs = dict()
        time_list = self.history('t')

        for var_key in var_keys:
            var_list = self.history(var_key)
            if var_list.ndim == 1:
                var_list = np.reshape(var_list, var_list.shape + (1, ))

            if var_key in var_ind_dict:
                ind = var_ind_dict[var_key]
            else:
                ind = list(range(var_list.shape[1]))

            if var_key in var_names_dict:
                names = var_names_dict[var_key]
            else:
                names = [var_key + "_" + str(k) for k in ind]

            subplot_num = len(ind)
            fig, ax = plt.subplots(subplot_num, 1)
            fig_axs[var_key] = {
                'fig': fig,
                'ax': ax
            }
            for i in range(subplot_num):
                if subplot_num == 1:
                    ax_ = ax
                else:
                    ax_ = ax[i]
                ax_.plot(time_list, var_list[:, ind[i]], label="Actual")
                ax_.set_xlabel("Time (s)")
                ax_.set_ylabel(names[i])
                ax_.grid()
                ax_.legend()
            fig.suptitle("Response of {:s} in {:s}".format(var_key, self.name))
            fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)
        return fig_axs

    def save(self, h5file=None, data_group=''):
        data_group = data_group + '/' + self.name
        self._logger.save(h5file, data_group)
        for sim_obj in self.sim_objs:
            sim_obj.save(h5file, data_group)

    def load(self, h5file=None, data_group=''):
        data_group = data_group + '/' + self.name
        self._logger.load(h5file, data_group)
        for sim_obj in self.sim_objs:
            sim_obj.load(h5file, data_group)


class StaticObject(SimObject):
    def __init__(self, eval_fun, interval: Union[int, float] = -1, **kwargs):
        super(StaticObject, self).__init__(interval=interval, **kwargs)
        self.eval_fun = eval_fun

    # implement
    def _forward(self, *args, **kwargs) -> Union[None, float, np.ndarray, dict]:
        return self.eval_fun(*args, **kwargs)


class DynSystem(SimObject):
    def __init__(self, initial_states: Dict[str, ArrayType], deriv_fun=None, output_fun=None, **kwargs):
        """ initial_states: dictionary of (state1, state2, ...) """
        super(DynSystem, self).__init__(interval=-1, **kwargs)
        self._add_state_vars(**initial_states)
        self.initial_states = initial_states

        if isinstance(deriv_fun, BaseFunction):
            self.deriv_fun = deriv_fun.evaluate
        else:
            self.deriv_fun = deriv_fun

        if isinstance(output_fun, BaseFunction):
            self.output_fun = output_fun.evaluate
        else:
            self.output_fun = output_fun

    # override
    def _reset(self):
        super(DynSystem, self)._reset()
        self.set_state(**self.initial_states)

    # may be implemented
    def _deriv(self, **kwargs) -> Dict[str, np.ndarray]:
        """ implement this method if needed
        args: dictionary of {name1: state1, name2: state2, ..., input_name1: input1, ...}
        return: dictionary of {name1: derivState1, name2: derivState2, ...)
        """
        if not self.deriv_fun:
            raise NotImplementedError
        return self.deriv_fun(**kwargs)

    # implement
    def _forward(self, **kwargs) -> Union[None, float, np.ndarray, dict]:
        states = self._get_states()
        derivs = self._deriv(**states, **kwargs)

        for name, var in self.state_vars.items():
            var.set_deriv(derivs[name])

        self._logger.append(t=self.time, **states, **kwargs)
        return self._output()

    # override
    @property
    def output(self) -> Union[None, float, np.ndarray, dict]:
        return self._output()

    # may be overridden
    def _output(self) -> Union[None, float, np.ndarray, dict]:
        if not self.output_fun:
            return None
        else:
            states = self._get_states()
            return self.output_fun(**states)


class TimeVaryingDynSystem(SimObject):
    def __init__(self, initial_states: Dict[str, ArrayType], deriv_fun=None, output_fun=None, **kwargs):
        super(TimeVaryingDynSystem, self).__init__(interval=-1, **kwargs)
        self._add_state_vars(**initial_states)
        self.initial_states = initial_states

        if isinstance(deriv_fun, BaseFunction):
            self.deriv_fun = deriv_fun.evaluate
        else:
            self.deriv_fun = deriv_fun

        if isinstance(output_fun, BaseFunction):
            self.output_fun = output_fun.evaluate
        else:
            self.output_fun = output_fun

    # override
    def _reset(self):
        super(TimeVaryingDynSystem, self)._reset()
        self.set_state(**self.initial_states)

    # to be implemented
    def _deriv(self, t, **kwargs) -> Dict[str, np.ndarray]:
        if self.deriv_fun is None:
            raise NotImplementedError
        return self.deriv_fun(t, **kwargs)

    # implement
    def _forward(self, **kwargs) -> Union[None, float, np.ndarray, dict]:
        states = self._get_states()
        derivs = self._deriv(t=self.time, **states, **kwargs)

        for name, var in self.state_vars.items():
            var.set_deriv(derivs[name])

        self._logger.append(t=self.time, **states, **kwargs)
        return self._output()

    # override
    @property
    def output(self) -> Union[None, float, np.ndarray, dict]:
        return self._output()

    # may be overridden
    def _output(self) -> Union[None, float, np.ndarray, dict]:
        if not self.output_fun:
            return None
        else:
            states = self._get_states()
            return self.output_fun(t=self.time, **states)


class BaseFunction(object):
    # to be implemented
    def evaluate(self, *args, **kwargs):
        raise NotImplementedError


if __name__ == "__main__":
    StateVariable.test()



"""
Description for Package
"""
from pysimenv.missile.model import PlanarKin, PitchDyn, PlanarVehicle, PlanarMissile,\
    PlanarMissileWithPitch, PlanarMovingTarget
from pysimenv.missile.engagement import Engagement2dim
from pysimenv.missile.guidance import PurePNG2dim, IACBPNG
from pysimenv.missile.control import ThreeLoopAP
from pysimenv.missile.util import RelKin2dim, CloseDistCond, miss_distance

__all__ = ['missile']